import os
import unittest.mock
import pandas as pd
import json
from typing import Optional
from collections import Counter

from medcat import cat
from medcat.data.model_card import ModelCard
from medcat.vocab import Vocab
from medcat.config import Config
from medcat.config.config_meta_cat import ConfigMetaCAT
from medcat.model_creation.cdb_maker import CDBMaker
from medcat.cdb import CDB
from medcat.tokenizing.tokens import UnregisteredDataPathException
from medcat.tokenizing.tokenizers import TOKENIZER_PREFIX
from medcat.utils.cdb_state import captured_state_cdb
from medcat.components.addons.meta_cat import MetaCATAddon
from medcat.utils.defaults import AVOID_LEGACY_CONVERSION_ENVIRON
from medcat.utils.defaults import LegacyConversionDisabledError

import unittest
import tempfile

from . import EXAMPLE_MODEL_PACK_ZIP
from . import V1_MODEL_PACK_PATH, UNPACKED_V1_MODEL_PACK_PATH
from .utils.legacy.test_conversion_all import ConvertedFunctionalityTests


orig_init = cat.CAT.__init__


class ModelLoadTests(unittest.TestCase):

    def assert_has_model_name(self, func):
        expected_model_packl_name = EXAMPLE_MODEL_PACK_ZIP.replace(".zip", "")

        def wrapper(*args, **kwargs):
            if 'model_load_path' in kwargs:
                self.assertEqual(kwargs['model_load_path'],
                                 expected_model_packl_name)
            else:
                self.assertEqual(args[-1], expected_model_packl_name)
            return func(*args, **kwargs)
        return wrapper

    def setUp(self):
        cat.CAT.__init__ = self.assert_has_model_name(cat.CAT.__init__)

    def tearDown(self):
        cat.CAT.__init__ = orig_init

    def test_loaded_model_knows_model_path(self):
        # NOTE: the assertion is checked due to wrapper on CAT.__init__
        inst = cat.CAT.load_model_pack(EXAMPLE_MODEL_PACK_ZIP)
        self.assertIsInstance(inst, cat.CAT)

    def test_can_load_CDB_from_model_pack(self):
        cdb = cat.CAT.load_cdb(EXAMPLE_MODEL_PACK_ZIP)
        self.assertIsInstance(cdb, CDB)


class TrainedModelTests(unittest.TestCase):
    TRAINED_MODEL_PATH = EXAMPLE_MODEL_PACK_ZIP

    @classmethod
    def setUpClass(cls):
        cls.model = cat.CAT.load_model_pack(cls.TRAINED_MODEL_PATH)
        if cls.model.config.components.linking.train:
            print("TRAINING WAS ENABLE! NEED TO DISABLE")
            cls.model.config.components.linking.train = False


class InferenceFromLoadedTests(TrainedModelTests):

    def test_can_load_model(self):
        self.assertIsInstance(self.model, cat.CAT)

    def test_has_training(self):
        self.assertTrue(self.model.cdb.cui2info)
        self.assertTrue(self.model.cdb.name2info)

    def test_inference_works(self):
        ents = self.model.get_entities(
            ConvertedFunctionalityTests.TEXT)['entities']
        for nr, ent in enumerate(ents.values()):
            with self.subTest(f"{nr}"):
                ConvertedFunctionalityTests.assert_has_ent(ent)

    def test_entities_in_correct_order(self):
        # NOTE: the issue wouldn't show up with smaller amount of text
        doc = self.model(ConvertedFunctionalityTests.TEXT * 3)
        cur_start = 0
        for ent in doc.linked_ents:
            with self.subTest(f"Ent: {ent}"):
                self.assertGreaterEqual(ent.base.start_char_index, cur_start)
                cur_start = ent.base.start_char_index


class CATIncludingTests(unittest.TestCase):
    TOKENIZING_PROVIDER = 'regex'
    EXPECT_TRAIN = {}

    # paths
    VOCAB_DATA_PATH = os.path.join(
        os.path.dirname(__file__), 'resources', 'vocab_data.txt'
    )
    CDB_PREPROCESSED_PATH = os.path.join(
        os.path.dirname(__file__), 'resources', 'preprocessed4cdb.txt'
    )

    @classmethod
    def setUpClass(cls):

        # vocab

        vocab = Vocab()
        vocab.add_words(cls.VOCAB_DATA_PATH)

        # CDB
        config = Config()

        # tokenizer
        config.general.nlp.provider = cls.TOKENIZING_PROVIDER

        maker = CDBMaker(config)

        cls.cdb: CDB = maker.prepare_csvs([cls.CDB_PREPROCESSED_PATH])

        # usage monitoring
        cls._temp_logs_folder = tempfile.TemporaryDirectory()
        config.general.usage_monitor.enabled = True
        config.general.usage_monitor.log_folder = cls._temp_logs_folder.name

        # CAT
        cls.cat = cat.CAT(cls.cdb, vocab)
        cls.cat.config.components.linking.train = False

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls._temp_logs_folder.cleanup()

    def tearDown(self):
        # remove existing contents / empty file log file
        log_file_path = self.cat.usage_monitor.log_file
        if os.path.exists(log_file_path):
            os.remove(log_file_path)


class CATCreationTests(CATIncludingTests):
    # should be persistent as long as we don't change the underlying model
    EXPECTED_HASH = "558019fd37ed2167"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.prev_hash = cls.cat.config.meta.hash

    @classmethod
    def get_cui2ct(cls, cat: Optional[cat.CAT] = None):
        if cat is None:
            cat = cls.cat
        return {
            cui: info['count_train'] for cui, info in cat.cdb.cui2info.items()
            if info['count_train']}

    def test_has_expected_training(self):
        self.assertEqual(self.get_cui2ct(), self.EXPECT_TRAIN)

    def test_versioning_updates_config_hash(self):
        self.assert_hashes_to(self.EXPECTED_HASH)

    def assert_hashes_to(self, exp_hash: str) -> None:
        self.cat._versioning(None)
        new_hash = self.cat.config.meta.hash
        self.assertNotEqual(self.prev_hash, new_hash)
        self.assertEqual(new_hash, exp_hash)
        self.assertEqual(self.cat.config.meta.history[-1], new_hash)

    def test_versioning_does_not_overpopulate_history(self):
        # run multiple times
        self.cat._versioning(None)
        self.cat._versioning(None)
        # and expect it not to append multiple times in the history
        # if there were multiple instances, the set would remove duplicates
        sorted_set = sorted(set(self.cat.config.meta.history))
        sorted_list = sorted(self.cat.config.meta.history)
        self.assertEqual(sorted_set, sorted_list)

    def test_can_get_model_card_str(self):
        model_card = self.cat.get_model_card(as_dict=False)
        self.assertIsInstance(model_card, str)

    def test_can_get_model_card_dict(self):
        model_card = self.cat.get_model_card(as_dict=True)
        self.assertIsInstance(model_card, dict)

    def test_model_card_has_required_keys(self):
        model_card = self.cat.get_model_card(as_dict=True)
        for ann in ModelCard.__annotations__:
            with self.subTest(f"Ann: {ann}"):
                self.assertIn(ann, model_card)

    def test_model_card_has_no_extra_keys(self):
        model_card = self.cat.get_model_card(as_dict=True)
        for key in model_card:
            with self.subTest(f"Key: {key}"):
                self.assertIn(key, ModelCard.__annotations__)


class CatWithMetaCATTests(CATCreationTests):
    EXPECTED_HASH = "8c3de3f171a87132"
    EXPECT_SAME_INSTANCES = True

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        meta_cat_cnf = ConfigMetaCAT()
        # NOTE: need to set for consistent hashing
        meta_cat_cnf.train.last_train_on = -1.0
        meta_cat_cnf.general.category_name = 'Status'
        meta_cat_cnf.general.tokenizer_name = 'bert-tokenizer'
        meta_cat_cnf.model.model_name = 'bert'
        meta_cat_cnf.model.model_variant = 'prajjwal1/bert-tiny'
        cls.addon = MetaCATAddon.create_new(
            meta_cat_cnf, cls.cat._pipeline.tokenizer)
        cls.cat.add_addon(cls.addon)
        cls.init_addons = list(cls.cat._pipeline._addons)

    def test_can_recreate_pipe(self):
        self.cat._recreate_pipe()
        addons_after = list(self.cat._pipeline._addons)
        self.assertGreater(len(self.init_addons), 0)
        self.assertEqual(len(self.init_addons), len(addons_after))
        if self.EXPECT_SAME_INSTANCES:
            self.assertEqual(self.init_addons, addons_after)
        else:
            # otherwise they should differ
            self.assertNotEqual(self.init_addons, addons_after)

    def test_get_entities_gets_monitored(self,
                                         text="Some text"):
        repeats = self.cat.config.general.usage_monitor.batch_size
        # ensure something gets written to the file
        for _ in range(repeats):
            self.cat.get_entities(text)
        log_file_path = self.cat.usage_monitor.log_file
        self.assertTrue(os.path.exists(log_file_path))
        with open(log_file_path) as f:
            contents = f.readline()
        self.assertTrue(contents)

    def test_get_entities_logs_usage(
            self,
            text="The dog is sitting outside the house."):
        # clear usage monitor buffer
        self.cat.usage_monitor.log_buffer.clear()
        self.cat.get_entities(text)
        self.assertTrue(self.cat.usage_monitor.log_buffer)
        self.assertEqual(len(self.cat.usage_monitor.log_buffer), 1)
        line = self.cat.usage_monitor.log_buffer[0]
        # the 1st element is the input text length
        input_text_length = line.split(",")[1]
        self.assertEqual(str(len(text)), input_text_length)


class CatWithMetaCATSaveLoadTests(CatWithMetaCATTests):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.temp_dir = tempfile.TemporaryDirectory()
        cls.mpp = cls.cat.save_model_pack(cls.temp_dir.name)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.temp_dir.cleanup()

    def test_can_save_pack(self):
        self.assertTrue(os.path.exists(self.mpp))

    def test_can_load_saved(self):
        loaded = cat.CAT.load_model_pack(self.mpp)
        self.assertIsInstance(loaded, cat.CAT)
        # test that it has an addon
        addons = list(loaded._pipeline.iter_addons())
        self.assertEqual(len(addons), 1)
        addon = addons[0]
        self.assertIsInstance(addon, MetaCATAddon)
        # test that loaded model pack has the same addon config as the addon
        self.assertIs(addon.config, loaded.config.components.addons[0])

    def test_can_load_meta_cat(self):
        addons = cat.CAT.load_addons(self.mpp)
        self.assertEqual(len(addons), 1)
        _, addon = addons[0]
        self.assertIsInstance(addon, MetaCATAddon)


class CatWithChangesMetaCATTests(CatWithMetaCATTests):
    EXPECTED_HASH = "0b22401059a08380"
    EXPECT_SAME_INSTANCES = False

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.addon.config.general.batch_size_eval = 10


class CATUnsupTrainingTests(CATCreationTests):
    SELF_SUPERVISED_DATA_PATH = os.path.join(
        os.path.dirname(__file__), 'resources', 'selfsupervised_data.txt'
    )
    EXPECT_TRAIN = {'C01': 2, 'C02': 2, 'C03': 2, 'C04': 1, 'C05': 1}
    # NOTE: should remain consistent unless we change the model or data
    EXPECTED_HASH = "e9989cc2dde739ff"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        data = pd.read_csv(cls.SELF_SUPERVISED_DATA_PATH)
        cls.cat.trainer.train_unsupervised(data.text.values)

    def test_lists_unsup_train_in_config(self):
        self.assertTrue(self.cat.config.meta.unsup_trained)


class CATSupTrainingTests(CATUnsupTrainingTests):
    SUPERVISED_DATA_PATH = os.path.join(
        os.path.dirname(__file__), 'resources', 'supervised_mct_export.json'
    )
    # NOTE: should remain consistent unless we change the model or data
    EXPECTED_HASH = "7bfe01e8e36eb07d"

    @classmethod
    def _get_cui_counts(cls) -> dict[str, int]:
        counter = Counter()
        data = cls._get_data()
        for proj in data['projects']:
            for doc in proj['documents']:
                for ann in doc['annotations']:
                    counter[ann['cui']] += 1
        return counter

    @classmethod
    def _get_data(cls) -> dict:
        with open(cls.SUPERVISED_DATA_PATH) as f:
            return json.load(f)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # copy from parent
        cls.EXPECT_TRAIN = CATUnsupTrainingTests.EXPECT_TRAIN.copy()
        cui_counts_in_data = cls._get_cui_counts()
        # add extra CUIs in supervised training example
        for cui, extra_cnt in cui_counts_in_data.items():
            cls.EXPECT_TRAIN[cui] += extra_cnt
        cls._perform_training()

    @classmethod
    def _perform_training(cls):
        data = cls._get_data()
        cls.cat.trainer.train_supervised_raw(data)

    def test_lists_sup_train_in_config(self):
        self.assertTrue(self.cat.config.meta.sup_trained)

    def test_clearing_training_works(self):
        with captured_state_cdb(self.cat.cdb):
            self.cat.cdb.reset_training()
            self.assertEqual(self.cat.cdb.get_cui2count_train(), {})
            self.assertEqual(self.cat.cdb.get_name2count_train(), {})
            self.assertEqual(self.cat.config.meta.unsup_trained, [])
            self.assertEqual(self.cat.config.meta.sup_trained, [])


class CATWithDictNERSupTrainingTests(CATSupTrainingTests):
    from medcat.components.types import CoreComponentType
    from medcat.components.ner.dict_based_ner import NER as DNER
    from medcat.components.ner.vocab_based_ner import NER as VNER

    @classmethod
    def _dummy_pt(cls):
        pass

    @classmethod
    def setUpClass(cls):
        # NOTE: need to do training AFTER changes
        #       so stopping it from happening here
        orig_training = cls._perform_training
        cls._perform_training = cls._dummy_pt
        super().setUpClass()
        cls._perform_training = orig_training
        cls.cdb.config.components.ner.comp_name = 'dict'
        cls.cat._recreate_pipe()
        # cls.cat.cdb.reset_training()
        cls._perform_training()

    def test_has_dict_based_ner(self):
        comp = self.cat._pipeline.get_component(self.CoreComponentType.ner)
        self.assertNotIsInstance(comp, self.VNER)
        self.assertIsInstance(comp, self.DNER)

    def test_can_get_entities(self,
                              expected_cuis: list[str] = ['C01', 'C05']):
        _ents = self.cat.get_entities(
            "The fittest most fit of chronic kidney failure", only_cui=True)
        ents = _ents['entities']
        self.assertEqual(len(ents), len(expected_cuis))
        self.assertEqual(set(ents.values()), set(expected_cuis))

    def test_can_get_multiple_entities(self):
        texts = [
            "The fittest most fit of chronic kidney failure",
            "The dog is sitting outside the house."
        ]
        ents = list(self.cat.get_entities_multi_texts(texts))
        self.assert_ents(ents, texts)

    def assert_ents(self, ents: list[tuple], texts: list[str]):
        self.assertEqual(len(ents), len(texts))
        # NOTE: text IDs are integers starting from 0
        exp_ids = set(str(i) for i in range(len(texts)))
        for ent_id_str, ent in ents:
            with self.subTest(f"Entity: {ent_id_str} [{ent}]"):
                self.assertIn(ent_id_str, exp_ids)

    def test_can_multiprocess_empty(self):
        texts = []
        ents = list(self.cat.get_entities_multi_texts(texts, n_process=3))
        self.assert_ents(ents, texts)

    def test_can_get_multiprocess(self):
        texts = [
            "The fittest most fit of chronic kidney failure",
            "The dog is sitting outside the house."
        ]
        ents = list(self.cat.get_entities_multi_texts(texts, n_process=3))
        self.assert_ents(ents, texts)


class CATWithDocAddonTests(CATIncludingTests):
    EXAMPLE_TEXT = "Example text to tokenize"
    ADDON_PATH = 'SMTH'
    EXAMPLE_VALUE = 'something else'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        doc = cls.cat(cls.EXAMPLE_TEXT)
        cls.doc_cls = doc.__class__
        cls.doc_cls.register_addon_path(cls.ADDON_PATH)
        # add for MutableEntity as well
        doc[:].register_addon_path(cls.ADDON_PATH)

    def setUp(self):
        self.doc = self.cat(self.EXAMPLE_TEXT)

    def test_can_set_value(self):
        self.doc.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)

    def test_cannot_set_incorrect_value(self):
        with self.assertRaises(UnregisteredDataPathException):
            self.doc.set_addon_data(self.ADDON_PATH * 2 + "#",
                                    self.EXAMPLE_TEXT)

    def test_cannot_get_incorrect_value(self):
        with self.assertRaises(UnregisteredDataPathException):
            self.doc.get_addon_data(self.ADDON_PATH * 2 + "#")

    def test_can_load_value(self):
        self.doc.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)
        got = self.doc.get_addon_data(self.ADDON_PATH)
        self.assertEqual(self.EXAMPLE_VALUE, got)

    def test_empty_doc_has_no_addon_data_paths(self):
        avail = self.doc.get_available_addon_paths()
        datas = {
            path: (data := self.doc.get_addon_data(path), bool(data),
                   self.doc.has_addon_data(path))
            for path in avail
        }
        self.assertFalse(avail, f"Available: {avail};\nDATA: {datas}")

    def test_doc_can_have_addon_data_path(self):
        # set some data
        self.doc.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)
        avail = self.doc.get_available_addon_paths()
        self.assertTrue(avail)
        datas = {
            path: (data := self.doc.get_addon_data(path), bool(data),
                   self.doc.has_addon_data(path))
            for path in avail
        }
        self.assertEqual(len(avail), 1, f"Available: {avail};\nDATA: {datas}")
        self.assertEqual(avail[0], self.ADDON_PATH)

    def test_empty_ent_has_no_addon_data_paths(self):
        ent = self.doc[:]
        avail = ent.get_available_addon_paths()
        self.assertFalse(avail)
        datas = {
            path: (data := self.doc.get_addon_data(path), bool(data),
                   self.doc.has_addon_data(path))
            for path in avail
        }
        self.assertFalse(ent.has_addon_data(self.ADDON_PATH),
                         f"Available: {avail};\nDATA: {datas}")

    def test_ent_can_have_addon_data_path(self):
        ent = self.doc[:]
        # set some data
        ent.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)
        avail = ent.get_available_addon_paths()
        self.assertTrue(avail)
        self.assertEqual(len(avail), 1)
        self.assertEqual(avail[0], self.ADDON_PATH)
        self.assertTrue(ent.has_addon_data(self.ADDON_PATH))


class MethodSpy:
    def __init__(self, obj: object, method_name: str):
        self.obj = obj
        self.is_module = obj.__class__.__name__ == 'module'
        self.method_name = method_name
        self.call_args = []
        self.call_results = []
        self._patcher = None
        self._original_method = getattr(obj, method_name)

    def __enter__(self):
        def wrapper(*args, **kwargs):
            self.call_args.append((args, kwargs))
            if self.is_module:
                result = self._original_method(*args, **kwargs)
            else:
                result = self._original_method(self.obj, *args, **kwargs)
            self.call_results.append(result)
            return result

        self._patcher = unittest.mock.patch.object(
            self.obj, self.method_name, side_effect=wrapper
        )
        self._patcher.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._patcher.stop()

    def assert_called_with(self, *args, **kwargs):
        assert (args, kwargs) in self.call_args, (
            f"No such args for, {self._original_method}")

    def assert_called_once_with(self, *args, **kwargs):
        assert len(self.call_args) == 1
        self.assert_called_with(
            *args, **kwargs)

    def assert_returned(self, ret_val):
        assert ret_val in self.call_results


class CATWithDocAddonSpacyTests(CATWithDocAddonTests):
    TOKENIZING_PROVIDER = 'spacy'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._save_folder = tempfile.TemporaryDirectory()
        cls.saved_model_path = cls.cat.save_model_pack(
            cls._save_folder.name, make_archive=False)
        # NOTE: that has changed config
        cls.saved_spacy_path = cls.cat.config.general.nlp.modelname

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls._save_folder.cleanup()

    def test_saves_spacy_model(self):
        # make sure was saved in the current folder
        self.assertIn(self._save_folder.name, self.saved_spacy_path)
        self.assertIn(TOKENIZER_PREFIX, self.saved_spacy_path)
        self.assertTrue(os.path.exists(self.saved_spacy_path))
        self.assertTrue(os.path.isdir(self.saved_spacy_path))

    def test_loads_spacy_model(self):
        import medcat.tokenizing.spacy_impl.tokenizers
        import spacy
        with MethodSpy(
                medcat.tokenizing.spacy_impl.tokenizers.SpacyTokenizer,
                "load_internals_from") as mock_load_internal:
            with MethodSpy(spacy, "load") as mock_load:
                model = cat.CAT.load_model_pack(self.saved_model_path)
        self.assertIsInstance(model, cat.CAT)
        mock_load_internal.assert_called_once_with(self.saved_spacy_path)
        mock_load.assert_called_once_with(
            self.saved_spacy_path,
            disable=self.cat.config.general.nlp.disabled_components)


class CATWithEntityAddonTests(CATIncludingTests):
    EXAMPLE_TEXT = "Example text to tokenize"
    EXAMPLE_ENT_START = 0
    EXAMPLE_ENT_END = 2
    ADDON_PATH = 'SMTH'
    EXAMPLE_VALUE = 'something else'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        doc = cls.cat(cls.EXAMPLE_TEXT)
        doc.__getitem__
        entity = doc[0:-1]
        cls.entity_cls = entity.__class__
        cls.entity_cls.register_addon_path(cls.ADDON_PATH)

    def setUp(self):
        self.doc = self.cat(self.EXAMPLE_TEXT)
        self.entity = self.doc[self.EXAMPLE_ENT_START: self.EXAMPLE_ENT_END]

    def test_can_add_data(self):
        self.entity.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)

    def test_cannot_add_data_to_wrong_path(self):
        with self.assertRaises(UnregisteredDataPathException):
            self.entity.set_addon_data(self.ADDON_PATH * 2 + "£",
                                       self.EXAMPLE_VALUE)

    def test_cannot_get_data_to_wrong_path(self):
        with self.assertRaises(UnregisteredDataPathException):
            self.entity.get_addon_data(self.ADDON_PATH * 2 + "£")

    def test_can_get_data(self):
        self.entity.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)
        got = self.entity.get_addon_data(self.ADDON_PATH)
        self.assertEqual(self.EXAMPLE_VALUE, got)

    def test_data_is_persistent(self):
        self.entity.set_addon_data(self.ADDON_PATH, self.EXAMPLE_VALUE)
        ent = self.doc[self.EXAMPLE_ENT_START: self.EXAMPLE_ENT_END]
        # new instance
        self.assertFalse(ent is self.entity)
        got = ent.get_addon_data(self.ADDON_PATH)
        self.assertEqual(self.EXAMPLE_VALUE, got)


class CATWithEntityAddonSpacyTests(CATWithEntityAddonTests):
    TOKENIZING_PROVIDER = 'spacy'


class CATLegacyLoadTests(unittest.TestCase):

    def test_can_load_legacy_model_zip(self):
        self.assertIsInstance(
            cat.CAT.load_model_pack(V1_MODEL_PACK_PATH), cat.CAT)

    def test_can_load_legacy_model_unpacked(self):
        self.assertIsInstance(
            cat.CAT.load_model_pack(UNPACKED_V1_MODEL_PACK_PATH), cat.CAT)

    def test_cannot_load_legacy_with_environ_set(self):
        with unittest.mock.patch.dict(os.environ, {
                AVOID_LEGACY_CONVERSION_ENVIRON: "true"}, clear=True):
            with self.assertRaises(LegacyConversionDisabledError):
                cat.CAT.load_model_pack(V1_MODEL_PACK_PATH)


class CATSaveTests(CATIncludingTests):
    DESCRIPTION = "Test CAT save functionality"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.temp_folder = tempfile.TemporaryDirectory()
        cls.saved_path = cls.cat.save_model_pack(
            cls.temp_folder.name, change_description=cls.DESCRIPTION)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.temp_folder.cleanup()

    def test_can_save_model_pack(self):
        self.assertTrue(os.path.exists(self.saved_path))

    def test_model_adds_description(self):
        self.assertIn(self.DESCRIPTION, self.cat.config.meta.description)


class BatchingTests(unittest.TestCase):
    NUM_TEXTS = 100
    all_texts = [
        f"Text {num:04d} -> " + "a" * num
        for num in range(NUM_TEXTS)
    ]
    total_text_length = sum(len(text) for text in all_texts)

    @classmethod
    def setUpClass(cls):
        cnf = Config()
        cls.cat = cat.CAT(cdb=CDB(cnf), vocab=Vocab())

    # per doc batching tests

    def test_batching_gets_full(self):
        batches = list(self.cat._generate_simple_batches(
            iter(self.all_texts), batch_size=self.NUM_TEXTS,
            only_cui=False))
        self.assertEqual(len(batches), 1)
        self.assertEqual(len(batches[0]), self.NUM_TEXTS)
        # NOTE: the contents has the text and the index and the only_cui bool
        #       so can't check equality directly
        # self.assertEqual(batches[0], self.all_texts)

    def test_batching_gets_in_sequence(self):
        batches = list(self.cat._generate_simple_batches(
            iter(self.all_texts), batch_size=self.NUM_TEXTS // 2,
            only_cui=False))
        self.assertEqual(len(batches), 2)
        self.assertEqual(len(batches[0]), self.NUM_TEXTS // 2)
        self.assertEqual(len(batches[1]), self.NUM_TEXTS // 2)
        # self.assertEqual(batches[0] + batches[1], self.all_texts)

    def test_batching_gets_all_1_at_a_time(self):
        batches = list(self.cat._generate_simple_batches(
            iter(self.all_texts), batch_size=1, only_cui=False))
        self.assertEqual(len(batches), self.NUM_TEXTS)
        for num, batch in enumerate(batches):
            with self.subTest(f"Batch {num}"):
                self.assertEqual(len(batch), 1)
                # self.assertEqual(batch[0], f"Text {num}")

    # per character batching tests

    def test_batching_gets_full_char(self):
        batches = list(self.cat._generate_batches_by_char_length(
            iter(self.all_texts), batch_size_chars=self.total_text_length,
            only_cui=False))
        self.assertEqual(len(batches), 1)
        # has all texts
        self.assertEqual(sum(len(batch) for batch in batches), self.NUM_TEXTS)
        # has all characters
        self.assertEqual(sum(len(text[0]) for text in batches[0]),
                         self.total_text_length)

    def test_batching_gets_all_half_at_a_time(self):
        exp_chars = int(0.7 * self.total_text_length)
        batches = list(self.cat._generate_batches_by_char_length(
            iter(self.all_texts), batch_size_chars=exp_chars,
            only_cui=False))
        # NOTE: should have 2 batches at 40% overlap
        self.assertEqual(len(batches), 2)
        # each batch should have less than expected characters
        for batch_num, batch in enumerate(batches):
            with self.subTest(f"Batch {batch_num}"):
                cur_total_chars = sum(len(text[1]) for text in batch)
                self.assertLessEqual(cur_total_chars, exp_chars)
        # has all texts
        self.assertEqual(sum(len(batch) for batch in batches), self.NUM_TEXTS)
        # has all characters
        self.assertEqual(sum(len(text[0])
                             for batch in batches for text in batch),
                         self.total_text_length)

    # overal batching (i.e joint methods)

    def test_cannot_set_both_neg(self):
        with self.assertRaises(ValueError):
            list(self.cat._generate_batches(
                iter(self.all_texts), batch_size_chars=-1,
                batch_size=-1, only_cui=False))

    def test_cannot_set_both_pos(self):
        with self.assertRaises(ValueError):
            list(self.cat._generate_batches(
                iter(self.all_texts), batch_size_chars=100,
                batch_size=10, only_cui=False))

    def test_can_do_char_based(self):
        exp_chars = int(0.3 * self.total_text_length)
        batches = list(self.cat._generate_batches(
            iter(self.all_texts), batch_size_chars=exp_chars,
            batch_size=-1, only_cui=False))
        self.assertGreater(len(batches), 0)
        batch_lens = [len(batch) for batch in batches]
        # has different number of texts in some batches -> not doc based
        self.assertGreater(max(batch_lens), min(batch_lens))

    def test_can_set_batch_size_per_doc(self):
        exp_batches = 10
        batches = list(self.cat._generate_batches(
            iter(self.all_texts), batch_size=exp_batches,
            batch_size_chars=-1, only_cui=False))
        self.assertGreater(len(batches), 0)
        batch_lens = [len(batch) for batch in batches]
        # has same number of texts in each batch -> doc based
        self.assertEqual(max(batch_lens), min(batch_lens))
        self.assertEqual(max(batch_lens), exp_batches)



import random
import logging
from typing import Iterator, Optional, Union

from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.tokenizing.tokens import MutableEntity, MutableDocument
from medcat.components.linking.vector_context_model import (
    ContextModel, PerDocumentTokenCache)
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import Config, ComponentConfig
from medcat.utils.defaults import StatusTypes as ST
from medcat.utils.postprocessing import create_main_ann
from medcat.tokenizing.tokenizers import BaseTokenizer


logger = logging.getLogger(__name__)


# class Linker(PipeRunner):
class Linker(AbstractCoreComponent):
    """Link to a biomedical database.

    Args:
        cdb (CDB): The Context Database.
        vocab (Vocab): The vocabulary.
        config (Config): The config.
    """

    # Custom pipeline component name
    name = 'medcat2_linker'

    # Override
    def __init__(self, cdb: CDB, vocab: Vocab, config: Config) -> None:
        self.cdb = cdb
        self.vocab = vocab
        self.config = config
        self.context_model = ContextModel(self.cdb.cui2info,
                                          self.cdb.name2info,
                                          self.cdb.weighted_average_function,
                                          self.vocab,
                                          self.config.components.linking,
                                          self.config.general.separator)
        # Counter for how often did a pair (name,cui) appear and
        # was used during training
        self.train_counter: dict = {}

    def get_type(self) -> CoreComponentType:
        return CoreComponentType.linking

    def _train(self, cui: str, entity: MutableEntity, doc: MutableDocument,
               per_doc_valid_token_cache: PerDocumentTokenCache,
               add_negative: bool = True) -> None:
        name = "{} - {}".format(entity.detected_name, cui)
        # TODO - bring back subsample after?
        # Always train
        self.context_model.train(
            cui, entity, doc, per_doc_valid_token_cache, negative=False)
        if (add_negative and
                self.config.components.linking.negative_probability
                >= random.random()):
            self.context_model.train_using_negative_sampling(cui)
        self.train_counter[name] = self.train_counter.get(name, 0) + 1

    def _process_entity_train(self, doc: MutableDocument,
                              entity: MutableEntity,
                              per_doc_valid_token_cache: PerDocumentTokenCache,
                              ) -> Iterator[MutableEntity]:
        cnf_l = self.config.components.linking
        # Check does it have a detected name
        if entity.detected_name is None:
            return
        name = entity.detected_name
        cuis = entity.link_candidates

        if len(name) < cnf_l.disamb_length_limit:
            return
        if len(cuis) == 1:
            # N - means name must be disambiguated, is not the preferred
            # name of the concept, links to other concepts also.
            name_info = self.cdb.name2info.get(name, None)
            if not name_info:
                return
            if name_info['per_cui_status'][cuis[0]] == ST.MUST_DISAMBIGATE:
                return
            self._train(cui=cuis[0], entity=entity, doc=doc,
                        per_doc_valid_token_cache=per_doc_valid_token_cache)
            entity.cui = cuis[0]
            entity.context_similarity = 1
            yield entity
        else:
            for cui in cuis:
                name_info = self.cdb.name2info.get(name, None)
                if not name_info:
                    continue
                if name_info['per_cui_status'][cui] not in ST.PRIMARY_STATUS:
                    continue
                # if self.cdb.name2cuis2status[name][cui] in {'P', 'PD'}:
                self._train(
                    cui=cui, entity=entity, doc=doc,
                    per_doc_valid_token_cache=per_doc_valid_token_cache)
                # It should not be possible that one name is 'P' for
                # two CUIs, but it can happen - and we do not care.
                entity.cui = cui
                entity.context_similarity = 1
                yield entity

    def _train_on_doc(self, doc: MutableDocument) -> Iterator[MutableEntity]:
        # Run training
        for entity in doc.ner_ents:
            yield from self._process_entity_train(
                doc, entity, PerDocumentTokenCache())

    def _process_entity_nt_w_name(
            self, doc: MutableDocument,
            entity: MutableEntity,
            cuis: list[str], name: str,
            per_doc_valid_token_cache: PerDocumentTokenCache
            ) -> tuple[Optional[str], float]:
        cnf_l = self.config.components.linking
        # NOTE: there used to be the condition
        # but if there are cuis, and it's an entity - surely, there's a match?
        # And there wasn't really an alternative anyway (which could have
        # caused and exception to be raised or cui/similarity from previous
        # entity to be used)
        # if len(cuis) > 0:
        do_disambiguate = False
        name_info = self.cdb.name2info[name]
        if len(name) < cnf_l.disamb_length_limit:
            do_disambiguate = True
        elif (len(cuis) == 1 and
                name_info['per_cui_status'][cuis[0]] in ST.DO_DISAMBUGATION):
            do_disambiguate = True
        elif len(cuis) > 1:
            do_disambiguate = True

        if do_disambiguate:
            cui, context_similarity = self.context_model.disambiguate(
                cuis, entity, name, doc, per_doc_valid_token_cache)
        else:
            cui = cuis[0]
            if self.config.components.linking.always_calculate_similarity:
                context_similarity = self.context_model.similarity(
                    cui, entity, doc, per_doc_valid_token_cache)
            else:
                context_similarity = 1  # Direct link, no care for similarity
        return cui, context_similarity

    def _check_similarity(self, cui: str, context_similarity: float) -> bool:
        th_type = self.config.components.linking.similarity_threshold_type
        threshold = self.config.components.linking.similarity_threshold
        if th_type == 'static':
            return context_similarity >= threshold
        if th_type == 'dynamic':
            conf = self.cdb.cui2info[cui]['average_confidence']
            return context_similarity >= conf * threshold
        return False

    def _process_entity_inference(
            self, doc: MutableDocument,
            entity: MutableEntity,
            per_doc_valid_token_cache: PerDocumentTokenCache
            ) -> Iterator[MutableEntity]:
        # Check does it have a detected concepts
        cuis = entity.link_candidates
        if not cuis:
            return
        # Check does it have a detected name
        name = entity.detected_name
        if name is not None:
            cui, context_similarity = self._process_entity_nt_w_name(
                doc, entity, cuis, name, per_doc_valid_token_cache)
        else:
            # No name detected, just disambiguate
            cui, context_similarity = self.context_model.disambiguate(
                cuis, entity, 'unk-unk', doc, per_doc_valid_token_cache)
        logger.debug("Considering CUI %s with sim %f",
                     cui, context_similarity)

        # Add the annotation if it exists and if above threshold and in filters
        cnf_l = self.config.components.linking
        if not cui or not cnf_l.filters.check_filters(cui):
            return
        if self._check_similarity(cui, context_similarity):
            entity.cui = cui
            entity.context_similarity = context_similarity
            yield entity

    def _inference(self, doc: MutableDocument) -> Iterator[MutableEntity]:
        per_doc_valid_token_cache = PerDocumentTokenCache()
        for entity in doc.ner_ents:
            logger.debug("Linker started with entity: %s", entity.base.text)
            yield from self._process_entity_inference(
                doc, entity, per_doc_valid_token_cache)

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        # Reset main entities, will be recreated later
        doc.linked_ents.clear()
        cnf_l = self.config.components.linking

        if cnf_l.train:
            linked_entities = self._train_on_doc(doc)
        else:
            linked_entities = self._inference(doc)
        # evaluating generator here because the `all_ents` list gets
        # cleared afterwards otherwise
        le = list(linked_entities)

        doc.ner_ents.clear()
        doc.ner_ents.extend(le)
        create_main_ann(doc)

        # TODO - reintroduce pretty labels? and apply here?

        # TODO - reintroduce groups? and map here?

        return doc

    def train(self, cui: str,
              entity: MutableEntity,
              doc: MutableDocument,
              negative: bool = False,
              names: Union[list[str], dict] = [],
              per_doc_valid_token_cache: Optional[PerDocumentTokenCache] = None
              ) -> None:
        """Train the linker.

        This simply trains the context model.

        Args:
            cui (str): The CUI to train.
            entity (BaseEntity): The entity we're at.
            doc (BaseDocument): The document within which we're working.
            negative (bool): Whether or not the example is negative.
                Defaults to False.
            names (list[str]/dict):
                Optionally used to update the `status` of a name-cui
                pair in the CDB.
            per_doc_valid_token_cache (PerDocumentTokenCache):
                Optionally, provide the per doc valid token cache.
        """
        if per_doc_valid_token_cache is None:
            per_doc_valid_token_cache = PerDocumentTokenCache()
        self.context_model.train(
            cui, entity, doc, per_doc_valid_token_cache, negative, names)

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'Linker':
        return cls(cdb, vocab, cdb.config)



from abc import ABC, abstractmethod
from typing import Optional, Union, overload
from tokenizers import Tokenizer

from medcat.config.config_meta_cat import ConfigMetaCAT


FAKE_TOKENIZER_PATH = "#\n/fake-path-not-exist#/"


class TokenizerWrapperBase(ABC):

    name: str

    def __init__(self, hf_tokenizer: Optional[Tokenizer] = None) -> None:
        self.hf_tokenizers = hf_tokenizer

    @overload
    def __call__(self, text: str) -> dict: ...

    @overload
    def __call__(self, text: list[str]) -> list[dict]: ...

    @abstractmethod
    def __call__(self, text: Union[str, list[str]]
                 ) -> Union[dict, list[dict]]: ...

    @abstractmethod
    def save(self, dir_path: str) -> None: ...

    @classmethod
    @abstractmethod
    def load(cls, dir_path: str, model_variant: Optional[str] = '', **kwargs
             ) -> Tokenizer: ...

    @abstractmethod
    def get_size(self) -> int: ...

    @abstractmethod
    def token_to_id(self, token: str) -> Union[int, list[int]]: ...

    @abstractmethod
    def get_pad_id(self) -> Union[Optional[int], list[int]]: ...

    def ensure_tokenizer(self) -> Tokenizer:
        if self.hf_tokenizers is None:
            raise ValueError("The tokenizer is not loaded yet")
        return self.hf_tokenizers


def init_tokenizer(cnf: ConfigMetaCAT) -> Optional[TokenizerWrapperBase]:
    tokenizer: Optional[TokenizerWrapperBase] = None
    if cnf.general.tokenizer_name == 'bbpe':
        from medcat.components.addons.meta_cat.mctokenizers.bpe_tokenizer import (  # noqa
            TokenizerWrapperBPE)
        tokenizer = TokenizerWrapperBPE.create_new()
    elif cnf.general.tokenizer_name == 'bert-tokenizer':
        from medcat.components.addons.meta_cat.mctokenizers.bert_tokenizer import (  # noqa
            TokenizerWrapperBERT)
        tokenizer = TokenizerWrapperBERT.create_new(cnf.model.model_variant)
    return tokenizer


def load_tokenizer(config: ConfigMetaCAT, tokenizer_folder: str
                   ) -> Optional[TokenizerWrapperBase]:
    tokenizer: Optional[TokenizerWrapperBase] = None
    if config.general.tokenizer_name == 'bbpe':
        from medcat.components.addons.meta_cat.mctokenizers.bpe_tokenizer import (  # noqa
            TokenizerWrapperBPE)
        tokenizer = TokenizerWrapperBPE.load(tokenizer_folder)
    elif config.general.tokenizer_name == 'bert-tokenizer':
        from medcat.components.addons.meta_cat.mctokenizers.bert_tokenizer import (  # noqa
            TokenizerWrapperBERT)
        tokenizer = TokenizerWrapperBERT.load(
            tokenizer_folder, config.model.model_variant)
    return tokenizer



"""De-identification model.

This describes a wrapper on the regular CAT model.
The idea is to simplify the use of a DeId-specific model.

It tackles two use cases
1) Creation of a deid model
2) Loading and use of a deid model

I.e for use case 1:

Instead of:
cat = CAT(cdb=ner.cdb, addl_ner=ner)

You can use:
deid = DeIdModel.create(ner)


And for use case 2:

Instead of:
cat = CAT.load_model_pack(model_pack_path)
anon_text = deid_text(cat, text)

You can use:
deid = DeIdModel.load_model_pack(model_pack_path)
anon_text = deid.deid_text(text)

Or if/when structured output is desired:
deid = DeIdModel.load_model_pack(model_pack_path)
anon_doc = deid(text)  # the spacy document

The wrapper also exposes some CAT parts directly:
- config
- cdb
"""
from typing import Union, Any, Optional, Iterable
import re
import logging

from medcat.cat import CAT
from medcat.cdb.cdb import CDB
from medcat.config.config_transformers_ner import ConfigTransformersNER
from medcat.components.types import CoreComponentType
from medcat.components.ner.trf.model import NerModel
from medcat.components.ner.trf.helpers import replace_entities_in_text
from medcat.components.ner.trf.transformers_ner import TransformersNER
from medcat.data.entities import Entity


logger = logging.getLogger(__name__)


class DeIdModel(NerModel):
    """The DeID model.

    This wraps a CAT instance and simplifies its use as a
    de-identification model.

    It provides methods for creating one from a TransformersNER
    as well as loading from a model pack (along with some validation).

    It also exposes some useful parts of the CAT it wraps such as
    the config and the concept database.
    """

    def __init__(self, cat: CAT) -> None:
        self.cat = cat

    def train(self, json_path: Union[str, list, None],
              *args, **kwargs) -> tuple[Any, Any, Any]:
        return super().train(json_path,
                             *args, **kwargs)  # type: ignore

    def deid_text(self, text: str, redact: bool = False) -> str:
        """Deidentify text and potentially redact information.

        De-identified text.
        If redaction is enabled, identifiable entities will be
        replaced with starts (e.g `*****`).
        Otherwise, the replacement will be the CUI or in other words,
        the type of information that was hidden (e.g [PATIENT]).

        Args:
            text (str): The text to deidentify.
            redact (bool): Whether to redact the information.

        Returns:
            str: The deidentified text.
        """
        entities = self.cat.get_entities(text)['entities']
        return replace_entities_in_text(text, entities, self.cat.cdb.get_name,
                                        redact=redact)

    def deid_multi_text(self, texts: Iterable[str], redact: bool = False,
                        n_process: Optional[int] = None) -> list[str]:
        if n_process is None:
            n_process = 1

        entities = self.cat.get_entities_multi_texts(
            texts, n_process=n_process)
        out: list[str] = []
        for raw_text, (_, _ents) in zip(texts, entities):
            ents = _ents['entities']
            text: str
            if isinstance(raw_text, tuple):
                text = raw_text[1]
            elif isinstance(raw_text, str):
                text = raw_text
            else:
                raise ValueError("Unknown raw text: "
                                 f"{type(raw_text)}: {raw_text}")
            new_text = replace_entities_in_text(
                text, ents, get_cui_name=self.cat.cdb.get_name, redact=redact)
            out.append(new_text)
        return out

    @classmethod
    def load_model_pack(cls, model_pack_path: str,
                        config: Optional[dict] = None) -> 'DeIdModel':
        """Load DeId model from model pack.

        The method first loads the CAT instance.

        It then makes sure that the model pack corresponds to a
        valid DeId model.

        Args:
            config: Config for DeId model pack (primarily for stride of
                overlap window)
            model_pack_path (str): The model pack path.

        Raises:
            ValueError: If the model pack does not correspond to a DeId model.

        Returns:
            DeIdModel: The resulting DeI model.
        """
        ner_model = NerModel.load_model_pack(model_pack_path, config=config)
        cat = ner_model.cat
        if not cls._is_deid_model(cat):
            raise ValueError(
                f"The model saved at {model_pack_path} is not a deid model "
                f"({cls._get_reason_not_deid(cat)})")
        model = cls(ner_model.cat)
        return model

    @classmethod
    def _is_deid_model(cls, cat: CAT) -> bool:
        return not bool(cls._get_reason_not_deid(cat))

    @classmethod
    def _get_reason_not_deid(cls, cat: CAT) -> str:
        if cat.vocab is not None:
            return "Has vocab"
        ner_comp = cat._pipeline.get_component(CoreComponentType.ner)
        if not isinstance(ner_comp, TransformersNER):
            raise ValueError(f"Incorrect NER component: {ner_comp.full_name}")
        return ""

    @classmethod
    def create(cls, cdb: CDB, cnf: ConfigTransformersNER):
        cdb.config.components.ner.comp_name = TransformersNER.name
        cdb.config.components.ner.custom_cnf = cnf
        # no-action linker
        cdb.config.components.linking.comp_name = 'no_action'
        cat = CAT(cdb=cdb, vocab=None, config=cdb.config)
        return cls(cat)


def match_rules(rules: list[tuple[str, str]], texts: list[str],
                cui2preferred_name: dict[str, str]) -> list[list[Entity]]:
    """Match a set of rules - pat / cui combos as post processing labels.
    Uses a cat DeID model for pretty name mapping.
    Args:
        rules (list[tuple[str, str]]): List of tuples of pattern and cui
        texts (list[str]): List of texts to match rules on
        cui2preferred_name (dict[str, str]): Dictionary of CUI to
            preferred name, likely to be cat.cdb.cui2preferred_name.
    Examples:
        >>> cat = CAT.load_model_pack(model_pack_path)
        ...
        >>> rules = [
            ('(123) 456-7890', '134'),
            ('1234567890', '134'),
            ('123.456.7890', '134'),
            ('1234567890', '134'),
            ('1234567890', '134'),
        ]
        >>> texts = [
            'My phone number is (123) 456-7890',
            'My phone number is 1234567890',
            'My phone number is 123.456.7890',
            'My phone number is 1234567890',
        ]
        >>> matches = match_rules(rules, texts, cat.cdb.cui2preferred_name)
    Returns:
        List[List[Dict]]: List of lists of predictions from `match_rules`
    """
    # Iterate through each text and pattern combination
    rule_matches_per_text: list[list[Entity]] = []
    for i, text in enumerate(texts):
        matches_in_text: list[Entity] = []
        for pattern, concept in rules:
            # Find all matches of current pattern in current text
            text_matches = re.finditer(pattern, text, flags=re.M)
            # Add each match with its pattern and text info
            for match in text_matches:
                matches_in_text.append({
                    'source_value': match.group(),
                    'pretty_name': cui2preferred_name[concept],
                    'start': match.start(),
                    'end': match.end(),
                    'cui': concept,
                    'acc': 1.0
                })
        rule_matches_per_text.append(matches_in_text)
    return rule_matches_per_text


def merge_all_preds(model_preds_by_text: list[list[Entity]],
                    rule_matches_per_text: list[list[Entity]],
                    accept_preds: bool = True) -> list[list[Entity]]:
    """Conveniance method to merge predictions from rule based and deID model
    predictions.

    Args:
        model_preds_by_text (list[list[Entity]]):
            List of predictions from `cat.get_entities()`, then
            `[list(m['entities'].values()) for m in model_preds]`
        rule_matches_per_text (list[list[Entity]]):
            List of predictions from output of running `match_rules`
        accept_preds (bool):
            Uses the predicted label from the model,
            model_preds_by_text, over the rule matches if they overlap.
            Defaults to using model preds over rules.
    Returns:
        list[list[Entity]]: List of lists of predictions from `merge_all_preds`
    """
    assert len(model_preds_by_text) == len(rule_matches_per_text), (
        "model_preds_by_text and rule_matches_per_text must have the same "
        "length as they should be CAT.get_entities and match_rules outputs of "
        "the same text")
    return [
        merge_preds(model_preds_by_text[i],
                    rule_matches_per_text[i],
                    accept_preds)
        for i in range(len(model_preds_by_text))]


def merge_preds(model_preds: list[Entity],
                rule_matches: list[Entity],
                accept_preds: bool = True) -> list[Entity]:
    """Merge predictions from rule based and deID model predictions.
    Args:
        model_preds (list[Entity]): predictions from `cat.get_entities()`
        rule_matches (list[Entity]): predictions from output of running
            `match_rules` on the same text
        accept_preds (bool): uses the predicted label from the model,
            model_preds, over the rule matches if they overlap.
            Defaults to using model preds over rules.
    Examples:
        >>> # a list of predictions from `cat.get_entities()`
        >>> model_preds = [
            [
                {'cui': '134', 'start': 10, 'end': 20, 'acc': 1.0,
                 'pretty_name': 'Phone Number'},
                {'cui': '134', 'start': 25, 'end': 35, 'acc': 1.0,
                 'pretty_name': 'Phone Number'}
            ]
        ]
        >>> # a list of predictions from `match_rules`
        >>> rule_matches = [
            [
                {'cui': '134', 'start': 10, 'end': 20, 'acc': 1.0,
                 'pretty_name': 'Phone Number'},
                {'cui': '134', 'start': 25, 'end': 35, 'acc': 1.0,
                 'pretty_name': 'Phone Number'}
            ]
        ]
        >>> merged_preds = merge_preds(model_preds, rule_matches)
    Returns:
        list[Entity]: List of predictions from `merge_preds`
    """
    if accept_preds:
        labels1 = model_preds
        labels2 = rule_matches
    else:
        labels1 = rule_matches
        labels2 = model_preds

    # Keep only non-overlapping model predictions
    labels2 = [span2 for span2 in labels2
               if not any(not (span2['end'] <= span1['start'] or
                               span1['end'] <= span2['start'])
                          for span1 in labels1)]
    # merge preds and sort on start
    merged_preds = labels1 + labels2
    merged_preds.sort(key=lambda x: x['start'])
    merged_preds
    return merged_preds



from typing import Iterator, Any, Optional, Union
from typing_extensions import TypedDict
from collections import Counter


class MedCATTrainerExportAnnotationRequired(TypedDict):
    start: int
    end: int
    cui: str
    value: str


class MedCATTrainerExportAnnotation(
        MedCATTrainerExportAnnotationRequired, total=False):
    id: Union[str, int]
    validated: Optional[bool]


class MedCATTrainerExportDocument(TypedDict):
    name: str
    id: Any
    last_modified: str
    text: str
    annotations: list[MedCATTrainerExportAnnotation]


class MedCATTrainerExportProject(TypedDict):
    name: str
    id: Any
    cuis: str
    tuis: Optional[str]
    documents: list[MedCATTrainerExportDocument]


MedCATTrainerExportProjectInfo = tuple[str, Any, str, Optional[str]]
"""The project name, project ID, CUIs str, and TUIs str"""


class MedCATTrainerExport(TypedDict):
    projects: list[MedCATTrainerExportProject]


def iter_projects(export: MedCATTrainerExport
                  ) -> Iterator[MedCATTrainerExportProject]:
    """Iterate over all the projects in the trainer export.

    Args:
        export (MedCATTrainerExport): The trainer export.

    Yields:
        Iterator[MedCATTrainerExportProject]: Project iterator.
    """
    yield from export['projects']


def iter_docs(export: MedCATTrainerExport
              ) -> Iterator[tuple[MedCATTrainerExportProjectInfo,
                                  MedCATTrainerExportDocument]]:
    """Iterate over all the docs in a trainer export.

    Args:
        export (MedCATTrainerExport): The trainer export.

    Yields:
        Iterator[tuple[MedCATTrainerExportProjectInfo,
                       MedCATTrainerExportDocument]]:
            The project info and the document.
    """
    for project in iter_projects(export):
        info: MedCATTrainerExportProjectInfo = (
            project['name'], project['id'], project['cuis'],
            project.get('tuis', None)
        )
        for doc in project['documents']:
            yield info, doc


def iter_anns(export: MedCATTrainerExport
              ) -> Iterator[tuple[MedCATTrainerExportProjectInfo,
                                  MedCATTrainerExportDocument,
                                  MedCATTrainerExportAnnotation]]:
    """Iterate over all the annotations in a trainer export.

    Args:
        export (MedCATTrainerExport): The trainer export.

    Yields:
        Iterator[tuple[MedCATTrainerExportProjectInfo,
                       MedCATTrainerExportDocument,
                       MedCATTrainerExportAnnotation]]:
            The project info, the document, and the annotation.
    """
    for proj_info, doc in iter_docs(export):
        for ann in doc['annotations']:
            yield proj_info, doc, ann


def count_all_annotations(export: MedCATTrainerExport) -> int:
    """Count the number of annotations in a trainer export.

    Args:
        export (MedCATTrainerExport): The trainer export.

    Returns:
        int: The total number of annotations.
    """
    return len(list(iter_anns(export)))


def count_all_docs(export: MedCATTrainerExport) -> int:
    """Count the number of documents in a trainer export.

    Args:
        export (MedCATTrainerExport): The trainer export.

    Returns:
        int: The total number of documents.
    """
    return len(list(iter_docs(export)))


def get_nr_of_annotations(doc: MedCATTrainerExportDocument) -> int:
    """Get the number of annotations for a tariner export document.

    Args:
        doc (MedCATTrainerExportDocument): The trainer export document.

    Returns:
        int: The number of annotations within the document.
    """
    return len(doc['annotations'])


def count_anns_per_concept(export: MedCATTrainerExport) -> dict[str, int]:
    counts: dict[str, int] = Counter()
    for _, _, ann in iter_anns(export):
        counts[ann['cui']] += 1
    return dict(counts)



from typing import Optional
import copy

from medcat.components.addons.meta_cat.mctokenizers.tokenizers import (
    TokenizerWrapperBase)
import logging

logger = logging.getLogger(__name__)


def prepare_from_json(data: dict,
                      cntx_left: int,
                      cntx_right: int,
                      tokenizer: TokenizerWrapperBase,
                      cui_filter: Optional[set] = None,
                      replace_center: Optional[str] = None,
                      prerequisites: dict = {},
                      lowercase: bool = True) -> dict:
    """Convert the data from a json format into a CSV-like format for
    training. This function is not very efficient (the one working with
    documents as part of the meta_cat.pipe method is much better).
    If your dataset is > 1M documents think about rewriting this function
    - but would be strange to have more than 1M manually annotated documents.

    Args:
        data (dict):
            Loaded output of MedCATtrainer. If we have a `my_export.json`
            from MedCATtrainer, than data = json.load(<my_export>).
        cntx_left (int):
            Size of context to get from the left of the concept
        cntx_right (int):
            Size of context to get from the right of the concept
        tokenizer (TokenizerWrapperBase):
            Something to split text into tokens for the LSTM/BERT/whatever
            meta models.
        replace_center (Optional[str]):
            If not None the center word (concept) will be replaced with
            whatever this is.
        prerequisites (dict):
            A map of prerequisites, for example our data has two
            meta-annotations (experiencer, negation). Assume I want to create
            a dataset for `negation` but only in those cases where
            `experiencer=patient`, my prerequisites would be:
                {'Experiencer': 'Patient'} - Take care that the CASE has to
                            match whatever is in the data. Defaults to `{}`.
        lowercase (bool):
            Should the text be lowercased before tokenization.
            Defaults to True.
        cui_filter (Optional[set]):
            CUI filter if set. Defaults to None.

    Returns:
        out_data (dict):
            Example: {'category_name': [('<category_value>', '<[tokens]>',
                        '<center_token>'), ...], ...}
    """
    out_data: dict = {}

    for project in data['projects']:
        for document in project['documents']:
            text = str(document['text'])
            if lowercase:
                text = text.lower()

            if len(text) > 0:
                doc_text = tokenizer(text)

                for ann in document.get('annotations', document.get(
                        # A hack to support entities and annotations
                        'entities', {}).values()):
                    cui = ann['cui']
                    skip = False
                    if 'meta_anns' in ann and prerequisites:
                        # It is possible to require certain meta_anns to exist
                        # and have a specific value
                        for meta_ann in prerequisites:
                            if (meta_ann not in ann['meta_anns'] or
                                    ann['meta_anns'][meta_ann][
                                        'value'] != prerequisites[meta_ann]):
                                # Skip this annotation as the prerequisite
                                # is not met
                                skip = True
                                break

                    if not skip and (cui_filter is None or
                                     not cui_filter or cui in cui_filter):
                        if ann.get('validated', True) and (
                                not ann.get('deleted', False) and
                                not ann.get('killed', False)
                                and not ann.get('irrelevant', False)):
                            start = ann['start']
                            end = ann['end']

                            # Updated implementation to extract all the tokens
                            # for the medical entity (rather than the one)
                            ctoken_idx = []
                            for ind, pair in enumerate(
                                    doc_text['offset_mapping']):
                                if start <= pair[0] or start <= pair[1]:
                                    if end <= pair[1]:
                                        ctoken_idx.append(ind)
                                        break
                                    else:
                                        ctoken_idx.append(ind)

                            _start = max(0, ctoken_idx[0] - cntx_left)
                            _end = min(len(doc_text['input_ids']),
                                       ctoken_idx[-1] + 1 + cntx_right)

                            cpos = cntx_left + min(0, ind - cntx_left)
                            cpos_new = [x - _start for x in ctoken_idx]
                            tkns = doc_text['input_ids'][_start:_end]

                            if replace_center is not None:
                                if lowercase:
                                    replace_center = replace_center.lower()
                                for p_ind, pair in enumerate(
                                        doc_text['offset_mapping']):
                                    if start >= pair[0] and start < pair[1]:
                                        s_ind = p_ind
                                    if end > pair[0] and end <= pair[1]:
                                        e_ind = p_ind

                                ln = e_ind - s_ind
                                tkns = tkns[:cpos] + tokenizer(
                                    replace_center)['input_ids'] + tkns[
                                        cpos + ln + 1:]

                            # Backward compatibility if meta_anns is a list vs
                            # dict in the new approach
                            meta_anns: list[dict] = []
                            if 'meta_anns' in ann:
                                if isinstance(ann['meta_anns'], dict):
                                    meta_anns.extend(ann['meta_anns'].values())
                                else:
                                    meta_anns.extend(ann['meta_anns'])

                            # If the annotation is validated
                            for meta_ann in meta_anns:
                                name = meta_ann['name']
                                value = meta_ann['value']

                                sample = [tkns, cpos_new, value]

                                if name in out_data:
                                    out_data[name].append(sample)
                                else:
                                    out_data[name] = [sample]
    return out_data


def prepare_for_oversampled_data(data: list,
                                 tokenizer: TokenizerWrapperBase) -> list:
    """Convert the data from a json format into a CSV-like format for
       training. This function is not very efficient (the one working with
       documents as part of the meta_cat.pipe method is much better).
       If your dataset is > 1M documents think about rewriting this function -
       but would be strange to have more than 1M manually annotated documents.

       Args:
           data (list):
               Oversampled data expected in the following format:
               [[['text','of','the','document'], [index of medical entity],
                    "label" ],
                ['text','of','the','document'], [index of medical entity],
                    "label" ]]
           tokenizer (TokenizerWrapperBase):
                Something to split text into tokens for the LSTM/BERT/whatever
                meta models.

       Returns:
            data_sampled (list):
                The processed data in the format that can be merged with the
                output from prepare_from_json.
                [[<[tokens]>, [index of medical entity], "label" ],
                <[tokens]>, [index of medical entity], "label" ]]
                """

    data_sampled = []
    for sample in data:
        # Checking if the input is already tokenized
        if isinstance(sample[0][0], str):
            doc_text = tokenizer(sample[0])
            data_sampled.append([
                doc_text[0]['input_ids'], sample[1], sample[2]])
        else:
            data_sampled.append([sample[0], sample[1], sample[2]])

    return data_sampled


def encode_category_values(data: dict,
                           existing_category_value2id: Optional[dict] = None,
                           category_undersample=None,
                           alternative_class_names: list[list[str]] = []
                           ) -> tuple:
    """Converts the category values in the data outputted by
    `prepare_from_json` into integer values.

    Args:
        data (dict):
            Output of `prepare_from_json`.
        existing_category_value2id(Optional[dict]):
            Map from category_value to id (old/existing).
        category_undersample:
            Name of class that should be used to undersample the data (for 2
            phase learning)
        alternative_class_names (list[list[str]]):
            A list of lists of strings, where each list contains variations
            of a class name. Usually read from the config at
            `config.general.alternative_class_names`.

    Returns:
        dict:
            New data with integers inplace of strings for category values.
        dict:
            New undersampled data (for 2 phase learning) with integers
            inplace of strings for category values
        dict:
            Map from category value to ID for all categories in the data.

    Raises:
        Exception: If categoryvalue2id is pre-defined and its labels do
            not match the labels found in the data
    """
    data_list = list(data)
    if existing_category_value2id is not None:
        category_value2id = existing_category_value2id
    else:
        category_value2id = {}

    category_values = set([x[2] for x in data_list])

    if (len(category_value2id) != 0 and
            set(category_value2id.keys()) != category_values):
        # if categoryvalue2id doesn't match the labels in the data,
        # then 'alternative_class_names' has to be defined to check
        # for variations
        if len(alternative_class_names) == 0:
            # Raise an exception since the labels don't match
            raise Exception(
                "The classes set in the config are not the same as the one "
                "found in the data. The classes present in the config vs the "
                "ones found in the data - {set(category_value2id.keys())}, "
                f"{category_values}. Additionally, ensure the populate the "
                "'alternative_class_names' attribute to accommodate for "
                "variations.")
        updated_category_value2id = {}
        for _class in category_value2id.keys():
            if _class in category_values:
                updated_category_value2id[_class] = category_value2id[_class]
            else:
                found_in = [sub_map for sub_map in alternative_class_names
                            if _class in sub_map]
                failed_to_find = False
                if len(found_in) != 0:
                    class_name_matched = [label for label in found_in[0]
                                          if label in category_values]
                    if len(class_name_matched) != 0:
                        updated_category_value2id[class_name_matched[0]
                                                  ] = category_value2id[_class]
                        logger.info(
                            "Class name '%s' does not exist in the data; "
                            "however a variation of it '%s' is present; "
                            "updating it...", _class, class_name_matched[0])
                    else:
                        failed_to_find = True
                else:
                    failed_to_find = True
                if failed_to_find:
                    raise Exception(
                        "The classes set in the config are not the same as "
                        "the one found in the data. The classes present in "
                        "the config vs the ones found in the data - "
                        f"{set(category_value2id.keys())}, {category_values}. "
                        "Additionally, ensure the populate the "
                        "'alternative_class_names' attribute to accommodate "
                        "for variations.")
        category_value2id = copy.deepcopy(updated_category_value2id)
        logger.info("Updated categoryvalue2id mapping - %s", category_value2id)
    # Else create the mapping from the labels found in the data
    else:
        for c in category_values:
            if c not in category_value2id:
                category_value2id[c] = len(category_value2id)
        logger.info("Categoryvalue2id mapping created with labels found "
                    "in the data - %s", category_value2id)

    # Map values to numbers
    for i in range(len(data_list)):
        data_list[i][2] = category_value2id[data_list[i][2]]

    # Creating dict with labels and its number of samples
    label_data_ = {v: 0 for v in category_value2id.values()}
    for i in range(len(data_list)):
        if data_list[i][2] in category_value2id.values():
            label_data_[data_list[i][2]] = label_data_[data_list[i][2]] + 1

    logger.info("Original number of samples per label: %s", label_data_)
    # Undersampling data
    if category_undersample is None or category_undersample == '':
        min_label = min(label_data_.values())

    else:
        if (category_undersample not in label_data_.keys() and
                category_undersample in category_value2id.keys()):
            min_label = label_data_[category_value2id[category_undersample]]
        else:
            min_label = label_data_[category_undersample]

    data_undersampled = []
    label_data_counter = {v: 0 for v in category_value2id.values()}

    for sample in data_list:
        if label_data_counter[sample[-1]] < min_label:
            data_undersampled.append(sample)
            label_data_counter[sample[-1]] += 1

    label_data = {v: 0 for v in category_value2id.values()}
    for i in range(len(data_undersampled)):
        if data_undersampled[i][2] in category_value2id.values():
            label_data[data_undersampled[i][2]] = label_data[
                data_undersampled[i][2]] + 1
    logger.info("Updated number of samples per label (for 2-phase learning): "
                "%s", label_data)

    return data_list, data_undersampled, category_value2id



import os
from datetime import datetime
from typing import Union, Callable
import platform
import logging
import sys

from medcat.config.config import UsageMonitor as UsageMonitorConfig


LOGS_ENV = "MEDCAT_USAGE_LOGS"
LOGS_LOC_ENV = "MEDCAT_USAGE_LOGS_LOCATION"

DEFAULT_LOGS_WINDOWS = os.path.join(os.environ.get('APPDATA', "NOT WINDOWS"),
                                    'medcat', 'logs')
DEFAULT_LOGS_LINUX = os.path.expanduser("~/.local/share/medcat/logs/")
DEFAULT_LOGS_MACOS = os.path.expanduser(
    "~/Library/Application Support/medcat/logs/")


logger = logging.getLogger(__name__)


class UsageMonitor:

    def __init__(self, model_hash: Union[Callable[[], str], str],
                 config: UsageMonitorConfig) -> None:
        self.config = config
        self.log_buffer: list[str] = []
        # NOTE: if the model hash changes (i.e model is trained)
        #       then this does not immediately take effect
        self._model_hash = model_hash

    @property
    def model_hash(self) -> str:
        return (self._model_hash() if callable(self._model_hash)
                else self._model_hash)

    @property
    def log_file(self):
        return os.path.join(
            self.config.log_folder,
            f"{self.config.file_prefix}{self.model_hash}.csv")

    def _get_auto_logs_location(self):
        system = platform.system().lower()
        if system == "windows":
            return DEFAULT_LOGS_WINDOWS
        elif system == "linux":
            return DEFAULT_LOGS_LINUX
        elif system == "darwin":  # macOS
            return DEFAULT_LOGS_MACOS
        else:
            raise OSError(f"Unsupported operating system: {system}")

    def _setup_auto_logs(self):
        # NOTE: os.environ is a snapshot of the environmental variables
        #       from the time that the process was started.
        #       However, someone could still change os.environm manually
        log_dir = os.environ.get(LOGS_LOC_ENV, self._get_auto_logs_location())
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        if log_dir != self.config.log_folder:
            self.config.log_folder = log_dir

    @property
    def should_monitor(self) -> bool:
        return self.config.enabled in (True, 'auto')

    def _should_log(self) -> bool:
        if not self.config.enabled:
            logger.warning("Trying to log to file when the usage monitor is "
                           "disabled. This should generally not happen unless "
                           "the config kept track of by the CAT object has is "
                           "different from the one kept track of by the usage "
                           "monitor. So if this keeps coming up, make sure "
                           "everything is up to date")
            return False
        elif self.config.enabled is True:
            return True
        elif self.config.enabled != 'auto':
            raise ValueError("Unknown UsageMonitor enabled status: "
                             f"{self.config.enabled}. Expected one of: "
                             f"True, False, 'auto'")
        # enabled == 'auto'
        env_enabled = os.environ.get(LOGS_ENV, "false").lower()
        if env_enabled in ("false", "0"):
            return False
        self._setup_auto_logs()
        return True

    def log_inference(self,
                      input_text_len: int,
                      nr_of_ents_found: int) -> None:
        if not self._should_log():
            return
        timestamp = datetime.now().isoformat()
        log_entry = f"{timestamp},{input_text_len},{nr_of_ents_found}"
        self.log_buffer.append(log_entry)
        if len(self.log_buffer) >= self.config.batch_size:
            self._flush_logs()

    def _flush_logs(self) -> None:
        if not self.log_buffer:
            return
        with open(self.log_file, 'a') as f:
            for log_entry in self.log_buffer:
                f.write(log_entry + '\n')
        self.log_buffer = []

    def __del__(self):
        # fail safe for when buffer is non-empty upon application stop
        # (i.e exit call)
        try:
            self._flush_logs()
        except FileNotFoundError:
            if not _in_test():
                raise
        except NameError as err:
            if not _says_open_not_available(err):
                raise


class _NoDelUM(UsageMonitor):

    def __del__(self):
        pass


# NOTE: at termination time, open is not available
#       while this isn't great, there's nothing we can do
def _says_open_not_available(err: NameError) -> bool:
    return "name 'open' is not defined" in str(err)


# NOTE: only relevant to flushing at deletion time
#       should not be used elsewhere
def _in_test() -> bool:
    return any(mod in sys.modules for mod in ("unittest", "pytest"))



import importlib.util
import importlib.metadata

import re

_DEP_PATTERN = re.compile(r"^([\w-]+)[<>=~;!].*")


class KeyDefaultDict(dict):
    def __missing__(self, key):
        return key


# Map the project name to the package needed to be imported where appropraite.
# Default to the package name itself.
_DEP_NAME_MAPPER = KeyDefaultDict({
    "pyahocorasick": "ahocorasick"
})


def get_all_extra_deps_raw(package_name: str) -> list[str]:
    """Get all the dependencies for a pcakge that are for an extra component.

    The output will include extra information such as the extra it's tied to.

    Args:
        package_name (str): The package name.

    Raises:
        ValueError: If the package isn't installed.

    Returns:
        list[str]: The list of extra dependencies, including extra information.
    """
    try:
        dependencies = importlib.metadata.requires(package_name)
        if dependencies is None:
            return []
    except importlib.metadata.PackageNotFoundError:
        raise ValueError(f"Package '{package_name}' is not installed")
    return [dep for dep in dependencies
            if "extra == " in dep]


def get_required_extra_deps(package_name: str, extra_name: str) -> list[str]:
    """Get the extra dependencies required for this extra part.

    Args:
        package_name (str): The package name.
        extra_name (str): The extra name.

    Returns:
        list[str]: All the required extra dependencies for this part.
    """
    dependencies = get_all_extra_deps_raw(package_name)
    return [
        # just package name
        _DEP_PATTERN.match(dep).group(1)  # type: ignore
        for dep in dependencies
        # check that this is an extra related to this name
        if (f"extra == '{extra_name}'" in dep
            or f'extra == "{extra_name}"' in dep)
    ]


def get_installed_extra_dependencies(package_name: str, extra_name: str
                                     ) -> list[str]:
    """Get installed dependencies for a given package's extra parts.

    Args:
        package_name (str): The package name.

    Returns:
        list[str]: The list of extra packages installed.
    """
    extra_deps = get_required_extra_deps(package_name, extra_name)
    return [dep for dep in extra_deps
            if importlib.util.find_spec(_DEP_NAME_MAPPER[dep]) is not None]


def ensure_optional_extras_installed(package_name: str, extra_name: str):
    """Ensure that an optional dependency set is installed.

    Args:
        package_name (str): The base package name.
        extra_name (str): The name of the extra dependency.

    Raises:
        MissingDependenciesError: If the extra dependency isn't provided.
    """
    installed = get_installed_extra_dependencies(package_name, extra_name)
    if not installed:
        req = get_required_extra_deps(package_name, extra_name)
        if not req:
            raise IncorrectExtraComponent(package_name, extra_name)
        missing = [requirement for requirement in req
                   if requirement not in installed]
        raise MissingDependenciesError(package_name, extra_name, missing)


class IncorrectExtraComponent(Exception):

    def __init__(self, package_name: str, extra_name: str):
        super().__init__(f"The '{extra_name}' part does not exist in as"
                         f"an optional part in {package_name} "
                         "(or does not define dependencies)")
        self.package_name = package_name
        self.extra_name = extra_name


class MissingDependenciesError(Exception):
    """Custom exception for missing optional dependencies."""

    def __init__(self, package_name: str, extra_name: str,
                 missing: list[str]):
        super().__init__(f"The optional dependency set '{extra_name}' "
                         f"is missing for {package_name}. The list of "
                         f"missing dependencies: {missing}. "
                         "If this came up when using the package, you need "
                         "to install the optional dependency alongside the "
                         "package. For instance, you can use "
                         f"`pip install {package_name}[{extra_name}]`. "
                         "PS: If you require multiple extras, you need to "
                         "specify them all ans separate by comma, e.g "
                         "`pkg[add1,add2]`")
        self.package_name = package_name
        self.extra_name = extra_name
        self.missing = missing



from typing import Any, Union, Optional

from medcat.components.ner.trf.transformers_ner import TransformersNER
from medcat.components.types import CoreComponentType
from medcat.tokenizing.tokens import MutableDocument
from medcat.cat import CAT
from medcat.cdb import CDB
from medcat.config import Config
from medcat.data.entities import Entities, OnlyCUIEntities


class NerModel:

    """The NER model.

    This wraps a CAT instance and simplifies its use as a
    NER model.

    It provides methods for creating one from a TransformersNER
    as well as loading from a model pack (along with some validation).

    It also exposes some useful parts of the CAT it wraps such as
    the config and the concept database.
    """

    def __init__(self, cat: CAT) -> None:
        self.cat = cat

    def train(self, json_path: Union[str, list, None],
              *args, **kwargs) -> tuple[Any, Any, Any]:
        """Train the underlying transformers NER model.

        All the extra arguments are passed to the TransformersNER train method.

        Args:
            json_path (Union[str, list, None]): The JSON file path to read the
                training data from.
            *args: Additional arguments for TransformersNER.train .
            **kwargs: Additional keyword arguments for TransformersNER.train .

        Returns:
            Tuple[Any, Any, Any]: df, examples, dataset
        """
        return self.trf_ner._component.train(json_path, *args, **kwargs)

    def eval(self, json_path: Union[str, list, None],
             *args, **kwargs) -> tuple[Any, Any, Any]:
        """Evaluate the underlying transformers NER model.
        All the extra arguments are passed to the TransformersNER eval method.
        Args:
            json_path (Union[str, list, None]):
                The JSON file path to read the training data from.
            *args: Additional arguments for TransformersNER.eval .
            **kwargs: Additional keyword arguments for TransformersNER.eval .
        Returns:
            Tuple[Any, Any, Any]: df, examples, dataset
        """
        return self.trf_ner._component.eval(json_path, *args, **kwargs)

    def __call__(self, text: Optional[str], *args, **kwargs
                 ) -> Optional[MutableDocument]:
        """Get the annotated document for text.

        Undefined arguments and keyword arguments get passed on to
        the equivalent `CAT` method.

        Args:
            text (Optional[str]): The input text.
            *args: Additional arguments for cat.__call__ .
            **kwargs: Additional keyword arguments for cat.__call__ .

        Returns:
            Optional[Doc]: The annotated document.
        """
        return self.cat(str(text), *args, **kwargs)

    def get_entities(self, text: str, *args, **kwargs
                     ) -> Union[dict, Entities, OnlyCUIEntities]:
        """Gets the entities recognized within a given text.

        The output format is identical to `CAT.get_entities`.

        Undefined arguments and keyword arguments get passed on to
        CAT.get_entities.

        Args:
            text (str): The input text.
            *args: Additional arguments for cat.get_entities .
            **kwargs: Additional keyword arguments for cat.get_entities .

        Returns:
            dict: The output entities.
        """
        return self.cat.get_entities(text, *args, **kwargs)

    def add_new_concepts(self,
                         cui2preferred_name: dict[str, str],
                         with_random_init: bool = False) -> None:
        """Add new concepts to the model and the concept database.

        Invoking this requires subsequent retraining on the model.

        Args:
            cui2preferred_name(dict[str, str]): Dictionary where each key is
                the literal ID of the concept to be added and each value is
                    its preferred name.
            with_random_init (bool): Whether to use the random init strategy
                for the new concepts. Defaults to False.
        """
        self.trf_ner._component.expand_model_with_concepts(
            cui2preferred_name, use_avg_init=not with_random_init)

    @property
    def trf_ner(self) -> TransformersNER:
        ner_comp = self.cat._pipeline.get_component(CoreComponentType.ner)
        if not isinstance(ner_comp, TransformersNER):
            raise ValueError(f"Incorrect NER component: {ner_comp.full_name}")
        return ner_comp

    @property
    def config(self) -> Config:
        return self.cat.config

    @property
    def cdb(self) -> CDB:
        return self.cat.cdb

    # @classmethod
    # def create(cls, ner: Union[TransformersNER, List[TransformersNER]]
    #            ) -> 'NerModel':
    #     """Create a NER model with a TransformersNER

    #     Args:
    #         ner (Union[TransformersNER, List[TransformersNER]]):
    #             The TransformersNER instance(s).

    #     Returns:
    #         NerModel: The resulting model
    #     """
    #     # expecting all to have the same CDB
    #     cdb = ner.cdb if isinstance(ner, TransformersNER) else ner[0].cdb
    #     cat = CAT(cdb=cdb, addl_ner=ner)
    #     return cls(cat)

    @classmethod
    def load_model_pack(cls, model_pack_path: str,
                        config: Optional[dict] = None) -> 'NerModel':
        """Load NER model from model pack.

        The method first wraps the loaded CAT instance.

        Args:
            config: Config for DeId model pack (primarily for stride of
                overlap window)
            model_pack_path (str): The model pack path.

        Returns:
            NerModel: The resulting DeI model.
        """
        cat = CAT.load_model_pack(model_pack_path)  # , ner_config_dict=config)
        return cls(cat)



import dill

from medcat.vocab import Vocab


def get_vocab_from_old(old_path: str) -> Vocab:
    """Convert a v1 vocab file to a v2 Vocab.

    Args:
        old_path (str): The v1 vocab file path.

    Returns:
        Vocab: The v2 Vocab.
    """
    with open(old_path, 'rb') as f:
        data = dill.load(f)
    v = Vocab()
    for word, word_data in data['vocab'].items():
        v.add_word(word, cnt=word_data['cnt'], vec=word_data['vec'],
                   replace=True)
    v.init_cumsums()
    return v



import os

EXPECTED_V1_CDB_FILE_NAME = "cdb.dat"
EXPECTED_V2_CDB_FOLDER_NAME = "cdb"


def is_legacy_model_pack(model_pack_path: str) -> bool:
    """Check if the model pack is a legacy model pack.

    Args:
        model_pack_path (str): The path to the model pack (unzipped).

    Returns:
        bool: True if the model pack is a legacy model pack, False otherwise.
    """
    if not os.path.isdir(model_pack_path):
        raise ValueError(
            f"Provided model pack path is not a directory: {model_pack_path}")
    cdb_path_v1 = os.path.join(model_pack_path, EXPECTED_V1_CDB_FILE_NAME)
    cdb_path_v2 = os.path.join(model_pack_path, EXPECTED_V2_CDB_FOLDER_NAME)
    return (
        # has cdb.dat and it is a file
        os.path.exists(cdb_path_v1) and os.path.isfile(cdb_path_v1) and
        # does not have cdb folder
        not os.path.exists(cdb_path_v2))



from typing import Any, Union, Protocol, runtime_checkable, Iterable
from enum import Enum, auto


class SerialisingStrategy(Enum):
    """Describes the strategy for serialising."""
    SERIALISABLE_ONLY = auto()
    """Only serialise attributes that are of Serialisable type"""
    SERIALISABLES_AND_DICT = auto()
    """Serialise attributes that are Serialisable as well as
    the rest of .__dict__"""
    DICT_ONLY = auto()
    """Only include the object's .__dict__"""
    MANUAL = auto()
    """Use manual serialisation defined by the object itself.

    NOTE: In this case, most of the logic defined within here will
          likely be ignored.
    """

    def _is_suitable_in_dict(self, attr_name: str,
                             attr: Any, obj: 'Serialisable') -> bool:
        if attr_name in obj.ignore_attrs():
            return False
        if self == SerialisingStrategy.SERIALISABLE_ONLY:
            return False
        elif self == SerialisingStrategy.DICT_ONLY:
            return True
        elif self == SerialisingStrategy.SERIALISABLES_AND_DICT:
            return not isinstance(attr, Serialisable)
        else:
            raise ValueError(f"Unknown instance: {self}")

    def _is_suitable_part(self, attr_name: str, part: Any, obj: 'Serialisable'
                          ) -> bool:
        if attr_name in obj.ignore_attrs():
            return False
        if not isinstance(part, Serialisable):
            return False
        if self == SerialisingStrategy.SERIALISABLE_ONLY:
            return True
        elif self == SerialisingStrategy.DICT_ONLY:
            return False
        return True

    def _iter_obj_items(self, obj: 'Serialisable'
                        ) -> Iterable[tuple[str, Any]]:
        for attr_name, attr in obj.__dict__.items():
            if attr_name.startswith("__"):
                # ignore privates
                continue
            yield attr_name, attr

    def _iter_obj_values(self, obj: 'Serialisable') -> Iterable[Any]:
        for _, val in self._iter_obj_items(obj):
            yield val

    def get_dict(self, obj: 'Serialisable') -> dict[str, Any]:
        """Gets the appropriate parts of the __dict__ of the object.

        I.e this filters out parts that shouldn't be included.

        Args:
            obj (Serialisable): The serialisable object.

        Returns:
            dict[str, Any]: The filtered attributes map.
        """
        out_dict = {
            attr_name: attr for attr_name, attr in self._iter_obj_items(obj)
            if self._is_suitable_in_dict(attr_name, attr, obj)
        }
        # do properties
        # NOTE: these are explicitly declared, so suitability is not checked
        out_dict.update({
            property_name: getattr(obj, property_name)
            for property_name in obj.include_properties()
        })
        return out_dict

    def get_parts(self, obj: 'Serialisable'
                  ) -> list[tuple['Serialisable', str]]:
        """Gets the matching serialisable parts of the object.

        This includes only serialisable parts, and only if specified
        by the strategy.

        Returns:
            list[tuple[Serialisable, str]]: The serialisable parts with names.
        """
        out_list: list[tuple[Serialisable, str]] = [
            (attr, attr_name) for attr_name, attr in self._iter_obj_items(obj)
            if self._is_suitable_part(attr_name, attr, obj)
        ]
        return out_list


@runtime_checkable
class Serialisable(Protocol):
    """The base serialisable protocol."""

    def get_strategy(self) -> SerialisingStrategy:
        """Get the serialisation strategy.

        Returns:
            SerialisingStrategy: The strategy.
        """
        pass

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        """Get the names of the arguments needed for init upon deserialisation.

        Returns:
            list[str]: The list of init arguments' names.
        """
        pass

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        """Get the names of attributes not to serialise.

        Returns:
            list[str]: The attribute names that should not be serialised.
        """
        pass

    @classmethod
    def include_properties(cls) -> list[str]:
        pass


class AbstractSerialisable:
    """The abstract serialisable base class.

    This defines some common defaults.
    """

    def get_strategy(self) -> SerialisingStrategy:
        return SerialisingStrategy.SERIALISABLES_AND_DICT

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return []

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return []

    @classmethod
    def include_properties(cls) -> list[str]:
        return []

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return False
        if set(self.__dict__) != set(other.__dict__):
            return False
        for attr_name, attr_value in self.__dict__.items():
            if not hasattr(other, attr_name):
                return False
            other_value = getattr(other, attr_name)
            if attr_value != other_value:
                return False
        return True


@runtime_checkable
class ManualSerialisable(Serialisable, Protocol):

    def serialise_to(self, folder_path: str) -> None:
        """Serialise to a folder.

        Args:
            folder_path (str): The folder to serialise to.
        """
        pass

    @classmethod
    def deserialise_from(cls, folder_path: str, **init_kwargs
                         ) -> 'ManualSerialisable':
        """Deserialise from a specifc path.

        The init keyword arguments are generally:
        - cnf: The config relevant to the components
        - tokenizer (BaseTokenizer): The base tokenizer for the model
        - cdb (CDB): The CDB for the model
        - vocab (Vocab): The Vocab for the model
        - model_load_path (Optional[str]): The model load path,
            but not the component load path

        Args:
            folder_path (str): The path to deserialsie form.

        Returns:
            ManualSerialisable: The deserialised object.
        """
        pass


class AbstractManualSerialisable:

    def get_strategy(self) -> SerialisingStrategy:
        return SerialisingStrategy.MANUAL

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return []

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return []

    @classmethod
    def include_properties(cls) -> list[str]:
        return []


def name_all_serialisable_elements(target_list: Union[list, tuple],
                                   name_start: str = '',
                                   all_or_nothing: bool = True
                                   ) -> list[tuple[Serialisable, str]]:
    """Gets all serialisable elements from a list or tuple.

    There's two strategies for finding the parts:
    1) If `all_or_nothing == True` either all the elements
        in the list must be Serialisable or None of them.
    2) If `all_or_nothing == False` some elements may be
        serialisable while others may not be.

    Args:
        target_list (Union[list, tuple]): The list/tuple of objects to look in.
        name_start (str, optional): The start of the name. Defaults to ''.
        all_or_nothing (bool, optional):
            Whether to disallow lists/tuple where only some elements are
            serialisable. Defaults to True.

    Raises:
        ValueError: If `all_or_nothing` is specified and not all elements
            are serialisable.

    Returns:
        list[tuple[Serialisable, str]]: The serialisable parts along with name.
    """
    out_parts: list[tuple[Serialisable, str]] = []
    if not target_list:
        return out_parts
    for el_nr, el in enumerate(target_list):
        if isinstance(el, Serialisable):
            out_parts.append((el, name_start + f"_el_{el_nr}"))
        elif all_or_nothing and out_parts:
            raise ValueError(f"The first {len(out_parts)} were serialisable "
                             "whereas the next one was not. Specify "
                             "`all_or_nothing=False` to allow for only "
                             "some of the list elements to be serialisable.")
    return out_parts


def get_all_serialisable_members(object: Serialisable
                                 ) -> tuple[list[tuple[Serialisable, str]],
                                            dict[str, Any]]:
    """Gets all serialisable members of an object.

    This looks for public and protected members, but not private ones.
    It should also be able to return parts of lists and tuples.
    It also provides the name of each serialisable object.

    Args:
        object (Any): The target object.

    Returns:
        tuple[list[tuple[Serialisable, str]], dict[str, Any]]:
            list of serialisable objects along with their names
    """
    strat = object.get_strategy()
    return strat.get_parts(object), strat.get_dict(object)



from typing import Iterator, Any, Type, Callable, Protocol, Union

import ast
import inspect
from enum import Enum
from functools import lru_cache
import random
import logging

from medcat.data.mctexport import (
    MedCATTrainerExport, MedCATTrainerExportDocument)


# this placheolder will be temporarily put in the
# phrases when dealing with one that has multiple
# of the same placeholder in it
_TEMP_MULTI_PLACEHOLDER = '###===PlaceHolder===###'


def partial_substitute(phrase: str, placeholder: str, name: str, nr: int
                       ) -> str:
    """Substitute all but 1 of the many placeholders present in the phrase.

    First, the first `nr` placeholders are replaced.
    Then the next (1) placeholder is replaced with a temporary one
    After that, the rest of the placeholders are replaced.
    And finally, the temporary placeholder is returned back to
    its original form.

    Example:
        If we've got `phrase = "some [PH] and [PH] we [PH]"`
        `placeholder = "[PH]"`, and `name = 'NAME'`,
        we'd get the following based on the number `nr`:
        0: "some [PH] and NAME we NAME"
        1: "some NAME and [PH] we NAME"
        2: "some NAME and NAME we [PH]"

    Args:
        phrase (str): The phrase in question.
        placeholder (str): The placeholder to replace.
        name (str): The name to replace the placeholder for.
        nr (int): The number of the target to keep.

    Raises:
        IncompatiblePhraseException: If the number of placeholders in the
            phrase is 1 or the number to be kept is too high; or the phrase
            has the temporary placeholder.

    Returns:
        str: The partially substituted phrase.
    """
    num_of_placeholder = phrase.count(placeholder)
    if nr >= num_of_placeholder or num_of_placeholder == 1:
        # NOTE: in cae of 1, this makes no sense
        raise IncompatiblePhraseException(
            f"The phrase ({repr(phrase)}) has {num_of_placeholder} "
            f"placeholders, but the {nr}th placeholder was requested to be "
            "swapped!")
    # replace stuff before the specific one
    phrase = phrase.replace(placeholder, name, nr)
    if _TEMP_MULTI_PLACEHOLDER in phrase:
        # if the temporary placeholder is already in text, the following
        # would fail unexpectedly
        raise IncompatiblePhraseException(
            f"Regression phrase with multiple placeholders ({placeholder}) "
            f"has the temporary placeholder: {repr(_TEMP_MULTI_PLACEHOLDER)}. "
            f"This means that the partial substitution of all but the {nr}th "
            "placeholder failed")
    # replace the target with temporary placeholder
    phrase = phrase.replace(placeholder, _TEMP_MULTI_PLACEHOLDER, 1)
    # replace the rest of the placeholder
    phrase = phrase.replace(placeholder, name)
    # set back the one needed placeholder
    phrase = phrase.replace(_TEMP_MULTI_PLACEHOLDER, placeholder)
    return phrase


class IncompatiblePhraseException(ValueError):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def limit_str_len(input_str: str,
                  max_length: int = 40,
                  keep_front: int = 20,
                  keep_rear: int = 10) -> str:
    """Limits the length of a string.

    If the length of the string is less than or equal to `max_length`, the same
    string is returned.
    If it's longer, the first `keep_front` are kept, then the number of chars
    is included in brackets (e.g `" [123 chars] "`), and finally the last
    `keep_rear` characters are included.

    Args:
        input_str (str): The input (potentially) long string.
        max_length (int): The maximum number of characters at which
            the string will remain unchanged. Defaults to 40.
        keep_front (int): How many starting characters to keep. Defaults to 20.
        keep_rear (int): How many ending characters to keep. Defaults to 10.

    Returns:
        str: _description_
    """
    if len(input_str) <= max_length:
        return input_str
    part1 = input_str[:keep_front]
    part2 = input_str[-keep_rear:]
    hidden_chars = len(input_str) - len(part1) - len(part2)
    return f"{part1} [{hidden_chars} chars] {part2}"


class MedCATTrainerExportConverter:
    """Used to convert an MCT export to the format required for regression."""
    # NOTE: the first placeholder will use the CUI, the 2nd the order of
    #       the annotation. This is required so that placeholders with the
    #       samme concept don't have the same name
    TEMP_PLACEHOLDER = "##[SWAPME-{}-{}]##"

    def __init__(self, mct_export: MedCATTrainerExport,
                 use_only_existing_name: bool = False) -> None:
        self.mct_export = mct_export
        self.use_only_existing_name = use_only_existing_name

    def _get_placeholder(self, cui: str, nr: int) -> str:
        return self.TEMP_PLACEHOLDER.format(cui, nr)

    def convert(self) -> dict:
        """Converts the MedCATtrainer export into regression suite dict.

        I.e this should producce a dict in the same format as one read
        from a regression suite YAML.

        Returns:
            dict: The Regression-suite compatible dict.
        """
        converted: dict[str, dict] = {}
        for phrase, case_name, anns in self._iter_docs():
            regr_case: dict[str, Any] = {
                'targeting': {
                    'placeholders': [
                        {
                            # NOTE: this is just and example.
                            #       it will be wiped/overwritten later
                            'placeholders': "TODO",
                            'cuis': ['CUI1']
                        }
                    ],
                    'any-combination': False,
                },
                'phrases': []  # will be filled later
            }
            placeholders: list[dict[str, Any]] = []
            # NOTE: the iteration is done from later annotations
            #       so I can replace using the locations
            for ann_nr, (start, end, cui, _) in enumerate(anns):
                ph = self._get_placeholder(cui, ann_nr)
                phrase = phrase[:start] + ph + phrase[end:]
                placeholders.append({
                    'placeholder': ph, 'cuis': [cui, ]
                })
            # update at the very end, when changed
            regr_case['phrases'] = [phrase]
            regr_case['targeting']['placeholders'] = placeholders
            converted[case_name] = regr_case
        return converted

    def _iter_docs(self) -> Iterator[
            tuple[str, str, Iterator[tuple[int, int, str, str]]]]:
        for project in self.mct_export['projects']:
            project_id = project['id']
            project_name = project['name']
            for doc in project['documents']:
                doc_id = doc['id']
                text = doc['text']
                yield (
                    text,
                    f"{project_id}_{project_name}_{doc_id}",
                    self._iter_anns_backwards(doc)
                )

    def _iter_anns_backwards(self, doc: MedCATTrainerExportDocument
                             ) -> Iterator[tuple[int, int, str, str]]:
        # NOTE: doing so backwards so that I can replace them one by one using
        #       the start/end, starting from the end of the phrase
        for ann in doc['annotations'][::-1]:
            yield ann['start'], ann['end'], ann['cui'], ann['value']


def get_class_level_docstrings(cls: Type) -> list[str]:
    """This is a helper method to get all the class level doc strings.

    This is designed to be used alongside and by the `add_doc_strings_to_enum`
    method.

    Args:
        cls (Type): The class in question.

    Returns:
        list[str]: All class-level docstrings.
            Includes the class docstring if it exists.
    """
    source_code = inspect.getsource(cls)
    tree = ast.parse(source_code)
    docstrings: list[str] = []
    # walk the tree
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            for elem in node.body:
                if isinstance(elem, ast.Expr) and isinstance(
                        elem.value, ast.Constant):
                    # If it's an expression node containing a constant,
                    # extract the string
                    docstrings.append(elem.value.s)
    return docstrings


def add_doc_strings_to_enum(cls: Type[Enum]) -> None:
    """Add doc strings to Enum as they are described in code right below
    each constant.

    The way python works means that the doc strings defined after an
    Enum constant do not get stored with the constant. When accessing
    the doc string of an Enum constant, the doc string of the class is
    returned instead.

    So what this method does is gets the doc strings by traversing the
    abstract syntax tree.

    While there would be easier ways to accomplish this, they would require
    the doc strings for the Enum constant to be further from the constants
    themselves.

    If the class itself has a doc string, it is omitted. Otherwise the Enum
    constants are given the doc strings in the order in which they appear.

    Args:
        cls (Type[Enum]): The Enum class to do this for.
    """
    docstrings = get_class_level_docstrings(cls)
    if cls.__doc__ == docstrings[0]:
        docstrings = docstrings[1:]
    for ev, ds in zip(cls, docstrings):
        ev.__doc__ = ds


@lru_cache(maxsize=10)
def get_rng(seed: int) -> random.Random:
    return random.Random(seed)


# NOTE: these are 'relatively accurate' estimates
#       that I obtained by running it on 15 different
#       concepts with names varying from length
#       of 5 to length of 74, a total of 316 names
ESTIMATION_MATRIX: dict[int, Callable[[int], int]] = {
    1: lambda orig_len: int(52.23 * orig_len + 24.26),
    2: lambda orig_len: 2724 * orig_len**2 + 3917 * orig_len + 1098
}


def estimate_num_variants(orig_len: int, edit_distance: int) -> int:
    if edit_distance in ESTIMATION_MATRIX:
        return ESTIMATION_MATRIX[edit_distance](orig_len)
    logging.warning("Estimations for then umber of varinats for edit "
                    "distance greater than 2 (%d used) can be extremely "
                    "inaccurate.")
    # NOTE: This is a low ball estimate - the real number could be a lot bigger
    powers = list(range(0, edit_distance + 1))[::-1]
    estimate_coefs = [(2 * 26) ** ed for ed in powers]
    estimated = 0
    for coef, power in zip(estimate_coefs, powers):
        estimated += coef * orig_len ** power
    return estimated


FAIL_AFTER_MULT = 10


def pick_random_edits(edit_gen: Union[list[str], set[str], Iterator[str]],
                      num_to_pick: int, orig_len: int, edit_distance: int,
                      rng_seed: int
                      ) -> Iterator[str]:
    num_vars = estimate_num_variants(orig_len, edit_distance)
    if num_to_pick > num_vars:
        raise ValueError(f"Unable to ick {num_to_pick} out of {num_vars} "
                         f"(estimated from edit distance {edit_distance} "
                         f"and word length {orig_len})")
    rng = get_rng(rng_seed)
    pick_avoids = num_to_pick > num_vars // 2
    _num_to_pick = num_to_pick if not pick_avoids else num_vars - num_to_pick
    pick_set: set[int] = set()
    while len(pick_set) < _num_to_pick:
        pick_set.add(rng.randint(0, num_vars))
    if pick_avoids:
        # NUMBERS NOT IN
        picks = sorted(set(range(num_vars)) - pick_set)
    else:
        picks = sorted(list(pick_set))
    for enr, edit in enumerate(edit_gen):
        if enr == picks[0]:
            picks.pop(0)
            yield edit
        if not picks:
            break


class EditGetter(Protocol):

    def __call__(self, word: str, use_diacritics: bool = False,
                 return_ordered: bool = False
                 ) -> Union[Iterator[str], set[str], list[str]]:
        pass



from typing import Type
import json
import logging


logger = logging.getLogger(__name__)


_CLASS_PATH = "serialised-class"
_INIT_PARTS_PATH = "init-parts"

# hidden file so that it doesn't get overwritten by other things
DEFAULT_SCHEMA_FILE = ".schema.json"


def _cls2path(cls: Type) -> str:
    return f"{cls.__module__}.{cls.__name__}"


def save_schema(file_name: str, cls: Type, init_parts: list[str]) -> None:
    """Saves the schema of a class to the specified file.

    Args:
        file_name (str): The file to save to.
        cls (Type): The class in question
        init_parts list[str]: The parts of the .
    """
    out_data = {
        _CLASS_PATH: _cls2path(cls),
        _INIT_PARTS_PATH: init_parts,
    }
    with open(file_name, 'w') as f:
        json.dump(out_data, f)


def load_schema(file_name: str) -> tuple[str, list[str]]:
    """Loads the schema for a folder of deserialisable files from the file.

    Args:
        file_name (str): The schema file

    Returns:
        tuple[str, list[str]]: The class package/name along
            with the parts needed for initialising.
    """
    with open(file_name) as f:
        data = json.load(f)
    class_name, init_parts = data[_CLASS_PATH], data[_INIT_PARTS_PATH]
    if __package__.startswith("medcat.") and class_name.startswith("medcat2."):
        # if we're loading a beta-release (medcat2.* namespaced) schema
        # we need to convert it to the stable one
        logger.info(
            "Loading beta-release medcat2 schema at '%s'. "
            "Converting to release schema.", file_name)
        class_name = class_name.replace("medcat2.", "medcat.")
    return class_name, init_parts


class IllegalSchemaException(ValueError):

    def __init__(self, *args):
        super().__init__("Illegal schema:", *args)



from typing import Optional, Iterable, Union, Sequence, cast, Callable
from typing import Protocol

import numpy as np
import random
import logging
from itertools import chain

from medcat.vocab import Vocab
from medcat.cdb.concepts import CUIInfo, NameInfo
from medcat.config.config import Linking
from medcat.tokenizing.tokens import (MutableToken, MutableEntity,
                                       MutableDocument)
from medcat.utils.defaults import StatusTypes as ST
from medcat.utils.matutils import unitvec
from medcat.storage.serialisables import AbstractSerialisable


logger = logging.getLogger(__name__)


class DisambPreprocessor(Protocol):

    def __call__(self, ent: MutableEntity, name: str, cuis: list[str],
                 similarities: list[float]) -> None:
        pass


class ContextModel(AbstractSerialisable):
    """Used to learn embeddings for concepts and calculate similarities
    in new documents.

    Args:
        cui2info (dict[str, CUIInfo]): The CUI to info mapping.
        name2info (dict[str, NameInfo]): The name to info mapping.
        weighted_average_function (Callable[[int], float]):
            The weighted average function.
        vocab (Vocab): The vocabulary
        config (Linking): The config to be used
        name_separator (str): The name separator
    """

    def __init__(self, cui2info: dict[str, CUIInfo],
                 name2info: dict[str, NameInfo],
                 weighted_average_function: Callable[[int], float],
                 vocab: Vocab, config: Linking,
                 name_separator: str,
                 disamb_preprocessors: list[DisambPreprocessor] = []) -> None:
        self.cui2info = cui2info
        self.name2info = name2info
        self.weighted_average_function = weighted_average_function
        self.vocab = vocab
        self.config = config
        self.name_separator = name_separator
        self._disamb_preprocessors = (  # copy if default/empty
            disamb_preprocessors or disamb_preprocessors.copy())

    def get_context_tokens(self, entity: MutableEntity, doc: MutableDocument,
                           size: int,
                           per_doc_valid_token_cache: 'PerDocumentTokenCache',
                           ) -> tuple[list[MutableToken],
                                      list[MutableToken],
                                      list[MutableToken]]:
        """Get context tokens for an entity, this will skip anything that
        is marked as skip in token._.to_skip

        Args:
            entity (BaseEntity): The entity to look for.
            doc (BaseDocument): The document look in.
            size (int): The size of the entity.
            per_doc_valid_token_cache (PerDocumentTokenCache):
                Per document cache for token validation.

        Returns:
            tuple[list[BaseToken], list[BaseToken], list[BaseToken]]:
                The tokens on the left, centre, and right.
        """
        start_ind = entity.base.start_index
        end_ind = entity.base.end_index

        _left_tokens = doc[max(0, start_ind - size):start_ind]
        tokens_left = [tkn for tkn in _left_tokens if
                       per_doc_valid_token_cache[tkn]]
        # Reverse because the first token should be the one closest to center
        tokens_left.reverse()
        tokens_center: list[MutableToken] = list(
            cast(Iterable[MutableToken], entity))
        _right_tokens = doc[end_ind + 1:end_ind + 1 + size]
        tokens_right = [tkn for tkn in _right_tokens if
                        per_doc_valid_token_cache[tkn]]

        return tokens_left, tokens_center, tokens_right

    def _tokens2vecs(self, tokens: Sequence[Union[MutableToken, str]]
                     ) -> Iterable[np.ndarray]:
        for step, tkn in enumerate(tokens):
            lower = tkn.lower() if isinstance(tkn, str) else tkn.base.lower
            if lower not in self.vocab:
                continue
            vec = self.vocab.vec(lower)
            if vec is not None:
                yield vec * self.weighted_average_function(step)

    def _should_change_name(self, cui: str) -> bool:
        target = self.config.random_replacement_unsupervised
        if random.random() <= target:
            return False
        return bool(self.cui2info.get(cui, None))

    def _preprocess_center_tokens(self, cui: Optional[str],
                                  tokens_center: list[MutableToken]
                                  ) -> Iterable[np.ndarray]:
        if cui is not None and self._should_change_name(cui):
            new_name: str = random.choice(list(self.cui2info[cui]['names']))
            new_tokens_center = new_name.split(self.name_separator)
            return self._tokens2vecs(new_tokens_center)
        else:
            return self._tokens2vecs(tokens_center)

    def get_context_vectors(self, entity: MutableEntity,
                            doc: MutableDocument,
                            per_doc_valid_token_cache: 'PerDocumentTokenCache',
                            cui: Optional[str] = None,
                            ) -> dict[str, np.ndarray]:
        """Given an entity and the document it will return the context
        representation for the given entity.

        Args:
            entity (BaseEntity): The entity to look for.
            doc (BaseDocument): The document to look in.
            per_doc_valid_token_cache (PerDocumentTokenCache):
                Per documnet cache for token validation
            cui (Optional[str]): The CUI or None if not specified.

        Returns:
            dict[str, np.ndarray]: The context vector.
        """
        vectors: dict[str, np.ndarray] = {}

        context_vector_sizes = self.config.context_vector_sizes
        for context_type, window_size in context_vector_sizes.items():
            tokens_left, tokens_center, tokens_right = self.get_context_tokens(
                entity, doc, window_size, per_doc_valid_token_cache)

            values: list[np.ndarray] = []
            # Add left
            values.extend(self._tokens2vecs(tokens_left))

            if not self.config.context_ignore_center_tokens:
                # Add center
                values.extend(
                    self._preprocess_center_tokens(cui, tokens_center))

            # Add right
            values.extend(self._tokens2vecs(tokens_right))

            if values:
                value = np.average(values, axis=0)
                vectors[context_type] = value
        return vectors

    def similarity(self, cui: str, entity: MutableEntity, doc: MutableDocument,
                   per_doc_valid_token_cache: 'PerDocumentTokenCache'
                   ) -> float:
        """Calculate the similarity between the learnt context for this CUI
        and the context in the given `doc`.

        Args:
            cui (str): The CUI.
            entity (BaseEntity): The entity to look for.
            doc (BaseDocument): The document to look in.
            per_doc_valid_token_cache (PerDocumentTokenCache):
                Per document cache for valid tokens

        Returns:
            float: The similarity.
        """
        vectors = self.get_context_vectors(
            entity, doc, per_doc_valid_token_cache)
        sim = self._similarity(cui, vectors)

        return sim

    def _similarity(self, cui: str, vectors: dict) -> float:
        """Calculate similarity once we have vectors and a cui.

        Args:
            cui (str): The CUI.
            vectors (dict): The vectors.

        Returns:
            float: The similarity.
        """
        cui_info = self.cui2info[cui]

        cui_vectors = cui_info['context_vectors']

        train_threshold = self.config.train_count_threshold
        if cui_vectors and cui_info['count_train'] >= train_threshold:
            return get_similarity(cui_vectors, vectors,
                                  self.config.context_vector_weights,
                                  cui, self.cui2info)
        else:
            return -1

    def _preprocess_disamb_similarities(self, entity: MutableEntity,
                                        name: str, cuis: list[str],
                                        similarities: list[float]) -> None:
        for preprocessor in self._disamb_preprocessors:
            preprocessor(entity, name, cuis, similarities)
        # NOTE: Has side effects on similarities
        if self.config.prefer_primary_name > 0:
            logger.debug("Preferring primary names")
            for i, (cui, sim) in enumerate(zip(cuis, similarities)):
                if sim <= 0:
                    continue
                status = self.name2info[name]['per_cui_status'].get(
                    cui, ST.AUTOMATIC)
                if status in ST.PRIMARY_STATUS:
                    new_sim = sim * (1 + self.config.prefer_primary_name)
                    similarities[i] = float(min(0.99, new_sim))
                    # DEBUG
                    logger.debug("CUI: %s, Name: %s, Old sim: %.3f, New "
                                 "sim: %.3f", cui, name, sim, similarities[i])

        if self.config.prefer_frequent_concepts > 0:
            logger.debug("Preferring frequent concepts")
            #  Prefer frequent concepts
            cnts = [self.cui2info[cui]['count_train'] for cui in cuis]
            m = min(cnts) or 1
            pref_freq = self.config.prefer_frequent_concepts
            scales = [np.log10(cnt / m) * pref_freq if cnt > 10 else 0
                      for cnt in cnts]
            old_sims = list(similarities)
            similarities.clear()
            similarities += [float(min(0.99, sim + sim * scale))
                             for sim, scale in zip(old_sims, scales)]

    def get_all_similarities(self, cuis: list[str], entity: MutableEntity,
                             name: str, doc: MutableDocument,
                             per_doc_valid_token_cache: 'PerDocumentTokenCache'
                             ) -> tuple[Union[list[str], list[None]],
                                        list[float], int]:
        vectors = self.get_context_vectors(
            entity, doc, per_doc_valid_token_cache)
        filters = self.config.filters

        # If it is trainer we want to filter concepts before disambiguation
        # do not want to explain why, but it is needed.
        if self.config.filter_before_disamb:
            # DEBUG
            logger.debug("Is trainer, subsetting CUIs")
            logger.debug("CUIs before: %s", cuis)

            cuis = [cui for cui in cuis if filters.check_filters(cui)]
            # DEBUG
            logger.debug("CUIs after: %s", cuis)

        if cuis:    # Maybe none are left after filtering
            # Calculate similarity for each cui
            similarities = [float(self._similarity(cui, vectors))
                            for cui in cuis]
            # DEBUG
            logger.debug("Similarities: %s", list(zip(cuis, similarities)))

            self._preprocess_disamb_similarities(
                entity, name, cuis, similarities)

            # technically, could be a np.int64 or something like that
            mx = int(np.argmax(similarities))
            return cuis, similarities, mx
        else:
            return [None], [0], 0

    def disambiguate(self, cuis: list[str], entity: MutableEntity, name: str,
                     doc: MutableDocument,
                     per_doc_valid_token_cache: 'PerDocumentTokenCache'
                     ) -> tuple[Optional[str], float]:
        suitable_cuis, sims, best_index = self.get_all_similarities(
            cuis, entity, name, doc, per_doc_valid_token_cache)
        return suitable_cuis[best_index], sims[best_index]

    def train(self, cui: str, entity: MutableEntity, doc: MutableDocument,
              per_doc_valid_token_cache: 'PerDocumentTokenCache',
              negative: bool = False, names: Union[list[str], dict] = [],
              ) -> None:
        """Update the context representation for this CUI, given it's correct
        location (entity) in a document (doc).

        Args:
            cui (str): The CUI to train.
            entity (BaseEntity): The entity we're at.
            doc (BaseDocument): The document within which we're working.
            per_doc_valid_token_cache (PerDocumentTokenCache):
                Per document cache for token validation.
            negative (bool): Whether or not the example is negative.
                Defaults to False.
            names (list[str]/dict):
                Optionally used to update the `status` of a name-cui
                pair in the CDB.
        """
        # Context vectors to be calculated
        if len(entity) == 0:  # Make sure there is something
            logger.warning("The provided entity for cui <%s> was empty, "
                           "nothing to train", cui)
            return
        vectors = self.get_context_vectors(
            entity, doc, per_doc_valid_token_cache, cui=cui)
        cui_info = self.cui2info[cui]
        lr = get_lr_linking(self.config, cui_info['count_train'])
        if not cui_info['context_vectors']:
            if not negative:
                cui_info['context_vectors'] = vectors
            else:
                cui_info['context_vectors'] = {ct: -1 * vec for
                                               ct, vec in vectors.items()}
        else:
            update_context_vectors(
                cui_info['context_vectors'], cui, vectors, lr,
                negative=negative)
        if not negative:
            cui_info['count_train'] += 1
        # Debug
        logger.debug("Updating CUI: %s with negative=%s", cui, negative)

        if not negative:
            # Update the name count, if possible
            if entity.detected_name:
                self.name2info[entity.detected_name]['count_train'] += 1

            if self.config.calculate_dynamic_threshold:
                # Update average confidence for this CUI
                sim = self.similarity(
                    cui, entity, doc, per_doc_valid_token_cache)
                new_conf = get_updated_average_confidence(
                    cui_info['average_confidence'],
                    cui_info['count_train'], sim)
                cui_info['average_confidence'] = new_conf

        if negative:
            # Change the status of the name so that it has
            # to be disambiguated always
            for name in names:
                if name not in self.name2info:
                    continue
                per_cui_status = self.name2info[name]['per_cui_status']
                cui_status = per_cui_status.get(cui, None)
                if cui_status == ST.PRIMARY_STATUS_NO_DISAMB:
                    # Set this name to always be disambiguated, even
                    # though it is primary
                    per_cui_status[cui] = ST.PRIMARY_STATUS_W_DISAMB
                    # Debug
                    logger.debug("Updating status for CUI: %s, "
                                 "name: %s to <%s>", cui, name,
                                 ST.PRIMARY_STATUS_W_DISAMB)
                elif cui_status == ST.AUTOMATIC:
                    # Set this name to always be disambiguated instead of A
                    per_cui_status[cui] = ST.MUST_DISAMBIGATE
                    logger.debug("Updating status for CUI: %s, "
                                 "name: %s to <N>", cui, name)
        if not negative and self.config.devalue_linked_concepts:
            # Find what other concepts can be disambiguated against this
            _other_cuis_chain = chain(*[
                self.name2info[name]['per_cui_status'].keys()
                for name in self.cui2info[cui]['names']])
            # Remove the cui of the current concept
            _other_cuis = set(_other_cuis_chain) - {cui}

            for _cui in _other_cuis:
                info = self.cui2info[_cui]
                if not info['context_vectors']:
                    info['context_vectors'] = vectors
                else:
                    update_context_vectors(
                        info['context_vectors'], cui, vectors, lr,
                        negative=True)

            logger.debug("Devalued via names.\n\tBase cui: %s \n\t"
                         "To be devalued: %s\n", cui, _other_cuis)

    def train_using_negative_sampling(self, cui: str) -> None:
        vectors = {}

        # Get vectors for each context type
        for context_type, size in self.config.context_vector_sizes.items():
            # While it should be size*2 it is already too many negative
            # examples, so we leave it at size
            ignore_pn = self.config.negative_ignore_punct_and_num
            inds = self.vocab.get_negative_samples(
                size, ignore_punct_and_num=ignore_pn)
            # NOTE: all indices in negative sampling have vectors
            #       since that's how they're generated
            values: list[np.ndarray] = self.vocab.get_vectors(inds)
            if len(values) > 0:
                vectors[context_type] = np.average(values, axis=0)
            # Debug
            logger.debug("Updating CUI: %s, with %s negative words",
                         cui, len(inds))

        cui_info = self.cui2info[cui]
        lr = get_lr_linking(self.config, cui_info['count_train'])
        # Do the update for all context types
        if not cui_info['context_vectors']:
            cui_info['context_vectors'] = vectors
        else:
            update_context_vectors(cui_info['context_vectors'], cui, vectors,
                                   lr, negative=True)


class PerDocumentTokenCache(dict[MutableToken, bool]):

    def __getitem__(self, key: MutableToken):
        index = key.base.index
        if index not in self:
            val = (not key.to_skip and not key.base.is_stop and
                   not key.base.is_digit and not key.is_punctuation)
            # NOTE: internally just using the token index
            self[index] = val  # type: ignore
            return val
        # NOTE: internally just using the token index
        return super().__getitem__(index)  # type: ignore


def get_lr_linking(config: Linking, cui_count: int) -> float:
    if config.optim['type'] == 'standard':
        return config.optim['lr']
    elif config.optim['type'] == 'linear':
        lr = config.optim['base_lr']
        cui_count += 1  # Just in case increase by 1
        return max(lr / cui_count, config.optim['min_lr'])
    else:
        raise Exception("Optimizer not implemented")


def get_similarity(cur_vectors: dict[str, np.ndarray],
                   other: dict[str, np.ndarray],
                   weights: dict[str, float], cui: str,
                   cui2info: dict[str, CUIInfo]) -> float:
    sim = 0
    for vec_type in weights:
        if vec_type not in other:
            # NOTE: sometimes the smaller context context types
            #       are unable to capture tokens that are present
            #       in our voab, which means they don't produce
            #       a value to be used here.
            continue
        if vec_type not in cur_vectors:
            # NOTE: this means that the saved vector doesn't have
            #       context at this vector type. This should be a
            #       rare occurrence, but is definitely present in
            #       models converted from v1
            continue
        w = weights[vec_type]
        v1 = cur_vectors[vec_type]
        v2 = other[vec_type]
        s = np.dot(unitvec(v1), unitvec(v2))
        sim += w * s
        logger.debug("Similarity for CUI: %s, Count: %s, Context Type: %.10s, "
                     "Weight: %s.2f, Similarity: %s.3f, S*W: %s.3f",
                     cui, cui2info[cui]['count_train'], vec_type, w, s, s * w)
    return float(sim)


def update_context_vectors(to_update: dict[str, np.ndarray], cui: str,
                           new_vecs: dict[str, np.ndarray], lr: float,
                           negative: bool) -> None:
    similarity = None
    for context_type, vector in new_vecs.items():
        # Get the right context
        if context_type in to_update:
            cv = to_update[context_type]
            similarity = np.dot(unitvec(cv), unitvec(vector))

            if negative:
                # Add negative context
                b = max(0, similarity) * lr
                to_update[context_type] = cv * (1 - b) - vector * b
            else:
                b = (1 - max(0, similarity)) * lr
                to_update[context_type] = cv * (1 - b) + vector * b

            # DEBUG
            logger.debug("Updated vector embedding.\n"
                         "CUI: %s, Context Type: %s, Similarity: %.2f, "
                         "Is Negative: %s, LR: %.5f, b: %.3f", cui,
                         context_type, similarity, negative, lr, b)
            cv = to_update[context_type]
            similarity_after = np.dot(unitvec(cv), unitvec(vector))
            logger.debug("Similarity before vs after: %.5f vs %.5f",
                         similarity, similarity_after)
        else:
            if negative:
                to_update[context_type] = -1 * vector
            else:
                to_update[context_type] = vector

            # DEBUG
            logger.debug("Added new context type with vectors.\n" +
                         "CUI: %s, Context Type: %s, Is Negative: %s",
                         cui, context_type, negative)


def get_updated_average_confidence(cur_ac: float, cnt_train: int,
                                   new_sim: float) -> float:
    return (cur_ac * cnt_train + new_sim) / (cnt_train + 1)



import logging
import os

from torch.optim import AdamW
from torch.optim.lr_scheduler import MultiStepLR

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.storage.serialisers import serialise
from medcat.components.addons.relation_extraction.models import (
    RelExtrBaseModel)
from medcat.components.addons.relation_extraction.pad_seq import (
    Pad_Sequence)
from medcat.components.addons.relation_extraction.tokenizer import (
    BaseTokenizerWrapper)
from medcat.components.addons.relation_extraction.config import (
    RelExtrBaseConfig)
from medcat.components.addons.relation_extraction.ml_utils import (
    load_state, save_state)

logger = logging.getLogger(__name__)


class RelExtrBaseComponent:

    name = "base_component_rel"

    def __init__(
            self,
            tokenizer: BaseTokenizerWrapper = BaseTokenizerWrapper(),
            model: RelExtrBaseModel = None,  # type: ignore
            model_config: RelExtrBaseConfig = None,  # type: ignore
            config: ConfigRelCAT = ConfigRelCAT(),
            task: str = "train",
            init_model: bool = False):
        """ Component that holds the model and everything for RelCAT.

        Args:
            tokenizer (BaseTokenizerWrapper): The base tokenizer for RelCAT.
            model (RelExtrBaseModel): The model wrapper.
            model_config (RelExtrBaseConfig):
                The model-specific config.
            config (ConfigRelCAT): The RelCAT config.
            task (str): The task - used for checkpointing.
            init_model (bool): Loads default BERT base model, tokenizer,
                model config. Defaults to False.
        """
        self.model: RelExtrBaseModel = model
        self.tokenizer: BaseTokenizerWrapper = tokenizer
        self.relcat_config: ConfigRelCAT = config
        self.model_config: RelExtrBaseConfig = model_config
        self.optimizer: AdamW = None  # type: ignore
        self.scheduler: MultiStepLR = None  # type: ignore
        self.task: str = task
        self.epoch: int = 0
        self.best_f1: float = 0.0

        if init_model:
            model_name = self.relcat_config.general.model_name
            self.model_config = RelExtrBaseConfig.load(
                pretrained_model_name_or_path=model_name,
                relcat_config=self.relcat_config)

            self.tokenizer = BaseTokenizerWrapper.load(
                tokenizer_path=model_name,
                relcat_config=self.relcat_config)

            special_tokens = self.relcat_config.general.tokenizer_relation_annotation_special_tokens_tags  # noqa
            self.tokenizer.hf_tokenizers.add_tokens(
                special_tokens,
                special_tokens=True)

            # used in llama tokenizer, may produce issues with other tokenizers
            self.tokenizer.hf_tokenizers.add_special_tokens(
                self.relcat_config.general.tokenizer_other_special_tokens)
            self.relcat_config.general.annotation_schema_tag_ids = (
                self.tokenizer.hf_tokenizers.convert_tokens_to_ids(
                    special_tokens))
            pad_idx = self.tokenizer.get_pad_id()
            self.relcat_config.model.padding_idx = pad_idx
            self.model_config.pad_token_id = pad_idx
            self.model_config.hf_model_config.vocab_size = (
                self.tokenizer.get_size())

            self.model = RelExtrBaseModel.load(
                pretrained_model_name_or_path=model_name,
                model_config=self.model_config,
                relcat_config=self.relcat_config)

            self.model.hf_model.resize_token_embeddings(
                self.tokenizer.get_size())

        self.pad_id = self.relcat_config.model.padding_idx
        self.padding_seq = Pad_Sequence(seq_pad_value=self.pad_id,
                                        label_pad_value=self.pad_id)

        logging.basicConfig(level=self.relcat_config.general.log_level)
        logger.setLevel(self.relcat_config.general.log_level)

        logger.info("RelExtrBaseComponent initialized")

    def save(self, save_path: str) -> None:
        """ Saves model and its dependencies to specified save_path folder.
            The CDB is obviously not saved, it is however necessary to save
            the tokenizer used.

        Args:
            save_path (str): folder path in which to save the model & deps.
        """

        assert self.relcat_config is not None
        cnf_path = os.path.join(save_path, "config")
        # NOTE: this'll be saved in a folder now
        os.makedirs(cnf_path, exist_ok=True)
        serialise('json', self.relcat_config, cnf_path)

        assert self.tokenizer is not None
        self.tokenizer.save(os.path.join(save_path))

        assert self.model is not None and self.model.hf_model is not None
        self.model.hf_model.resize_token_embeddings(self.tokenizer.get_size())

        assert self.model_config is not None
        self.model_config.hf_model_config.vocab_size = (
            self.tokenizer.get_size())
        self.model_config.hf_model_config.pad_token_id = self.pad_id
        self.model_config.save(save_path)

        save_state(self.model, optimizer=self.optimizer,
                   scheduler=self.scheduler, epoch=self.epoch,
                   best_f1=self.best_f1, path=save_path,
                   model_name=self.relcat_config.general.model_name,
                   task=self.task, is_checkpoint=False, final_export=True)

    @classmethod
    def load(cls, pretrained_model_name_or_path: str = "./"
             ) -> "RelExtrBaseComponent":
        """
        Args:
            pretrained_model_name_or_path (str): Path to RelCAT model.
                Defaults to "./".

        Returns:
            RelExtrBaseComponent: component.
        """

        relcat_config = ConfigRelCAT.load(
            load_path=pretrained_model_name_or_path)
        return cls.from_relcat_config(
            relcat_config, pretrained_model_name_or_path)

    @classmethod
    def from_relcat_config(cls, relcat_config: ConfigRelCAT,
                           pretrained_model_name_or_path: str = './'
                           ) -> 'RelExtrBaseComponent':
        model_config = RelExtrBaseConfig.load(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config)

        tokenizer = BaseTokenizerWrapper.load(
            tokenizer_path=pretrained_model_name_or_path,
            relcat_config=relcat_config)

        model = RelExtrBaseModel.load(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            model_config=model_config, relcat_config=relcat_config)

        model.hf_model.resize_token_embeddings(len(tokenizer.hf_tokenizers))

        optimizer = None
        scheduler = None

        epoch, best_f1 = load_state(
            model, optimizer, scheduler, path=pretrained_model_name_or_path,
            model_name=relcat_config.general.model_name,
            file_prefix=relcat_config.general.task,
            relcat_config=relcat_config)

        component = cls(model=model, tokenizer=tokenizer,
                        model_config=model_config, config=relcat_config)
        cls.epoch = epoch
        cls.best_f1 = best_f1

        return component



from typing import Protocol, Type, Callable, runtime_checkable
from typing_extensions import Self
import logging

from medcat.config import Config
from medcat.tokenizing.tokens import (MutableDocument, MutableEntity,
                                      MutableToken)
from medcat.utils.registry import Registry


logger = logging.getLogger(__name__)


TOKENIZER_PREFIX = "tokenizer_internals_"


class BaseTokenizer(Protocol):
    """The base tokenizer protocol."""

    def create_entity(self, doc: MutableDocument,
                      token_start_index: int, token_end_index: int,
                      label: str) -> MutableEntity:
        """Create an entity from a document.

        Args:
            doc (MutableDocument): The document to use.
            token_start_index (int): The token start index.
            token_end_index (int): The token end index.
            label (str): The label.

        Returns:
            MutableEntity: The resulting entity.
        """
        pass

    def entity_from_tokens(self, tokens: list[MutableToken]) -> MutableEntity:
        """Get an entity from the list of tokens.

        Args:
            tokens (list[MutableToken]): List of tokens.

        Returns:
            MutableEntity: The resulting entity.
        """
        pass

    def __call__(self, text: str) -> MutableDocument:
        pass

    @classmethod
    def create_new_tokenizer(cls, config: Config) -> Self:
        pass

    def get_doc_class(self) -> Type[MutableDocument]:
        """Get the document implementation class used by the tokenizer.

        This can be used (e.g) to register addon paths.

        Returns:
            Type[MutableDocument]: The document class.
        """
        pass

    def get_entity_class(self) -> Type[MutableEntity]:
        """Get the entity implementation class used by the tokenizer.

        Returns:
            Type[MutableEntity]: The entity class.
        """
        pass


@runtime_checkable
class SaveableTokenizer(Protocol):

    def save_internals_to(self, folder_path: str) -> str:
        """Save tokenizer internals to specified folder.

        The returning folder's basename should start with `TOKENIZER_PREFIX`.

        Args:
            folder_path (str): The folder to use for the internals.

        Returns:
            str: The subfolder the internals were saved to.
        """

    def load_internals_from(self, folder_path: str) -> bool:
        """Attempt to load internals from a folder path.

        If the specified folder exists, internals will be loaded.
        If the folder doesn't exist, nothing will be loaded.

        The given folder's basename should start with `TOKENIZER_PREFIX`.

        Args:
            folder_path (str): The path to the folder to load internals from.

        Returns:
            bool: Whether the loading was successful.
        """
        pass


_DEFAULT_TOKENIZING: dict[str, tuple[str, str]] = {
    "regex": ("medcat.tokenizing.regex_impl.tokenizer",
              "RegexTokenizer.create_new_tokenizer"),
    "spacy": ("medcat.tokenizing.spacy_impl.tokenizers",
              "SpacyTokenizer.create_new_tokenizer")
}

_TOKENIZERS_REGISTRY = Registry(BaseTokenizer,  # type: ignore
                                lazy_defaults=_DEFAULT_TOKENIZING)


def get_tokenizer_creator(tokenizer_name: str
                          ) -> Callable[[Config], BaseTokenizer]:
    """Get the creator method for the tokenizer.

    While this is generally just the class instance (i.e refers
    to the `___init__`), another callable can be used internally.

    Args:
        tokenizer_name (str): The name of the tokenizer.

    Returns:
        Callable[[Config], BaseTokenizer]: The creator for the tokenizer.
    """
    return _TOKENIZERS_REGISTRY.get_component(tokenizer_name)


def create_tokenizer(tokenizer_name: str, config: Config) -> BaseTokenizer:
    """Create the tokenizer given the init arguments.

    Args:
        tokenizer_name (str): The tokenizer name.
        config (Config): The config to be passed to the constructor.

    Returns:
        BaseTokenizer: The created tokenizer.
    """
    return _TOKENIZERS_REGISTRY.get_component(tokenizer_name)(config)


def list_available_tokenizers() -> list[tuple[str, str]]:
    """Get the available tokenizers.

    Returns:
        list[tuple[str, str]]: The list of the name, and class name
            of the available tokenizer.
    """
    return _TOKENIZERS_REGISTRY.list_components()


def register_tokenizer(name: str, clazz: Type[BaseTokenizer]) -> None:
    """Register a new tokenizer.

    Args:
        name (str): The name of the tokenizer.
        clazz (Type[BaseTokenizer]): The class of the tokenizer (i.e creator).
    """
    _TOKENIZERS_REGISTRY.register(name, clazz)
    logger.debug("Registered tokenizer '%s': '%s.%s'",
                 name, clazz.__module__, clazz.__name__)



from typing import Optional

from medcat.config.config import SerialisableBaseModel
from medcat.utils.hasher import Hasher


class General(SerialisableBaseModel):
    """The general part of the Transformers NER config"""
    name: str = 'deid'
    model_name: str = 'roberta-base'
    """Can be path also"""
    seed: int = 13
    description: str = "No description"
    """Should provide a basic description of this MetaCAT model"""
    pipe_batch_size_in_chars: int = 20000000
    """How many characters are piped at once into the meta_cat class"""
    ner_aggregation_strategy: str = 'simple'
    """Agg strategy for HF pipeline for NER"""
    chunking_overlap_window: Optional[int] = 5
    """Size of the overlap window used for chunking"""
    test_size: float = 0.2
    last_train_on: Optional[float] = None
    verbose_metrics: bool = False

    class Config:
        extra = 'allow'
        validate_assignment = True
        protected_namespaces = ()


class ConfigTransformersNER(SerialisableBaseModel):
    """The transformer NER config"""
    general: General = General()

    class Config:
        extra = 'allow'
        validate_assignment = True

    def get_hash(self, hasher: Optional[Hasher] = None) -> str:
        if hasher is None:
            hasher = Hasher()
        for v in self.model_dump().values():
            hasher.update(v)
        return hasher.hexdigest()



from typing import Optional
import re

from medcat.config.config import Preprocessing
from medcat.components.types import CoreComponentType, AbstractCoreComponent
from medcat.tokenizing.tokens import MutableDocument
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import ComponentConfig


class TagAndSkipTagger(AbstractCoreComponent):
    name = 'tag-and-skip-tagger'

    def __init__(self, preprocessing: Preprocessing) -> None:
        self.word_skipper = re.compile('^({})$'.format(
            '|'.join(preprocessing.words_to_skip)))
        # Very aggressive punct checker, input will be lowercased
        self.punct_checker = re.compile(r'[^a-z0-9]+')
        self.cnf_p = preprocessing

    def get_type(self) -> CoreComponentType:
        return CoreComponentType.tagging

    def __call__(self, doc: MutableDocument) -> MutableDocument:
        for token in doc:
            if (self.punct_checker.match(token.base.lower) and
                    token.base.text not in self.cnf_p.keep_punct):
                # There can't be punct in a token if it also has text
                token.is_punctuation = True
                token.to_skip = True
            elif self.word_skipper.match(token.base.lower):
                # Skip if specific strings
                token.to_skip = True
            elif self.cnf_p.skip_stopwords and token.base.is_stop:
                token.to_skip = True

        return doc


    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'TagAndSkipTagger':
        return cls(cdb.config.preprocessing)



import re
from dataclasses import dataclass
from typing import Protocol

from medcat.tokenizing.tokens import MutableDocument
from medcat.tokenizing.tokenizers import BaseTokenizer


@dataclass
class NameDescriptor:
    tokens: list[str]
    snames: set[str]
    raw_name: str
    is_upper: bool


class LGeneral(Protocol):
    separator: str


class LPreprocessing(Protocol):
    min_len_normalize: int
    do_not_normalize: set[str]


class LCDBMaker(Protocol):
    name_versions: list[str]
    min_letters_required: int


def _get_tokens(config: LPreprocessing,
                sc_name: MutableDocument,
                version: str) -> list[str]:
    if version == "LOWER":
        return [tkn.base.lower for tkn in sc_name if not tkn.to_skip]
    if version == "CLEAN":
        min_norm_len = config.min_len_normalize
        tokens = []
        for tkn in sc_name:
            if not tkn.to_skip:
                if len(tkn.base.lower) < min_norm_len:
                    tokens.append(tkn.base.lower)
                elif (config.do_not_normalize and
                        tkn.tag is not None and
                        tkn.tag in config.do_not_normalize):
                    tokens.append(tkn.base.lower)
                else:
                    tokens.append(tkn.lemma.lower())
        return tokens
    raise UnknownTokenVersion(version)


def _update_dict(configs: tuple[LGeneral, LPreprocessing, LCDBMaker],
                 raw_name: str,
                 names: dict[str, NameDescriptor],
                 tokens: list[str], is_upper: bool) -> None:
    general, _, cdb_maker = configs
    snames = set()
    name = general.separator.join(tokens)
    mlr = cdb_maker.min_letters_required
    if (mlr and len(re.sub("[^A-Za-z]*", '', name)) < mlr):
        return  # too short
    if name in names:
        return  # already exists
    sname = ""
    for token in tokens:
        if sname:
            sname = sname + general.separator + token
        else:
            sname = token
        snames.add(sname.strip())

    names[name] = NameDescriptor(tokens=tokens, snames=snames,
                                 raw_name=raw_name, is_upper=is_upper)


def prepare_name(raw_name: str, nlp: BaseTokenizer,
                 names: dict[str, NameDescriptor],
                 configs: tuple[LGeneral, LPreprocessing, LCDBMaker],
                 ) -> dict[str, NameDescriptor]:
    """Generates different forms of a name. Will edit the provided `names`
    dictionary and add information generated from the `name`.

    Args:
        nlp (BaseTokenizer): The tokenizer.
        names (dict[str, NameDescriptor]):
            Dictionary of existing names for this concept in this row of a CSV.
            The new generated name versions and other required information will
            be added here.
        configs (tuple[LGeneral, LPreprocessing, LCDBMaker]):
            Applicable configs for medcat.

    Returns:
        names (dict):
            The updated dictionary of prepared names.
    """
    sc_name = nlp(raw_name)
    _, preprocessing, cdb_maker = configs

    for version in cdb_maker.name_versions:
        tokens = None

        tokens = _get_tokens(preprocessing, sc_name, version)

        if tokens is not None and tokens:
            _update_dict(configs, raw_name, names, tokens,
                         sc_name.base.isupper())

    return names


class UnknownTokenVersion(ValueError):

    def __init__(self, version: str) -> None:
        super().__init__(f"Unknown token version: '{version}'")
